from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import HDFS_APACHE_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthenticationMethod(Enum):
    kerberos = "kerberos"
    password = "password"


class FilesystemType(Enum):
    httpfs = "httpfs"
    webhdfs = "webhdfs"


class HdfsApacheConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "c10e5224-f17d-4524-844f-e97b1305e489"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    authentication_method: HDFS_APACHE_CONNECTION.AuthenticationMethod | None = Field(
        HDFS_APACHE_CONNECTION.AuthenticationMethod.password, alias="authentication_method"
    )
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    connect_to_apache_hive: bool | None = Field(False, alias="connect_to_apache_hive")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    hive_database: str | None = Field(None, alias="hive_db")
    hive_host: str | None = Field(None, alias="hive_host")
    hive_http_path: str | None = Field(None, alias="hive_http_path")
    hive_keytab_file: str | None = Field(None, alias="hive_keytab")
    hive_password: str | None = Field(None, alias="hive_password")
    hive_port: int | None = Field(None, alias="hive_port")
    hive_service_principal_name: str | None = Field(None, alias="hive_service_principal")
    enable_ssl_for_hive: bool | None = Field(True, alias="hive_ssl")
    hive_username: str | None = Field(None, alias="hive_user")
    hive_user_principal_name: str | None = Field(None, alias="hive_user_principal")
    keytab_file: str | None = Field(None, alias="keytab")
    password: str | None = Field(None, alias="password")
    service_principal_name: str | None = Field(None, alias="service_principal")
    ssl_certificate: str | None = Field(None, alias="ssl_certificate")
    url: str = Field(None, alias="url")
    use_home_as_root: bool | None = Field(True, alias="use_home_as_root")
    user_principal_name: str = Field(None, alias="user_principal")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    secure_gateway_id: str | None = Field(None, alias="sg_gateway_id")
    sg_host_original: str | None = Field(None, alias="sg_host_original")
    secure_gateway_as_http_proxy: bool | None = Field(None, alias="sg_http_proxy")
    secure_gateway_security_token: str | None = Field(None, alias="sg_security_token")
    secure_gateway_service_url: str | None = Field(None, alias="sg_service_url")
    satellite_client_certificate: str | None = Field(None, alias="sl_client_cert")
    satellite_client_private_key: str | None = Field(None, alias="sl_client_private_key")
    satellite_connector_id: str | None = Field(None, alias="sl_connector_id")
    satellite_endpoint_host: str | None = Field(None, alias="sl_endpoint_host")
    satellite_endpoint_display_name: str | None = Field(None, alias="sl_endpoint_name")
    satellite_endpoint_port: int | None = Field(None, alias="sl_endpoint_port")
    sl_host_original: str | None = Field(None, alias="sl_host_original")
    satellite_as_http_proxy: bool | None = Field(None, alias="sl_http_proxy")
    satellite_location_id: str | None = Field(None, alias="sl_location_id")
    satellite_service_url: str | None = Field(None, alias="sl_service_url")
    file_system_type: HDFS_APACHE_CONNECTION.FilesystemType | None = Field(None, alias="filesystem_type")
    hive_login_config_name: str | None = Field(None, alias="hive_login_config_name")
    ssl_certificate_file: str | None = Field(None, alias="ssl_certificate_file")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("hive_password")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value != "kerberos"
                        )
                        or (self.authentication_method != "kerberos")
                    )
                )
                and (self.connect_to_apache_hive)
            )
            else exclude.add("hive_password")
        )
        include.add("hive_port") if (self.connect_to_apache_hive) else exclude.add("hive_port")
        include.add("ssl_certificate") if (not self.ssl_certificate_file) else exclude.add("ssl_certificate")
        include.add("hive_http_path") if (self.connect_to_apache_hive) else exclude.add("hive_http_path")
        (
            include.add("keytab_file")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
            )
            else exclude.add("keytab_file")
        )
        (
            include.add("hive_username")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value != "kerberos"
                        )
                        or (self.authentication_method != "kerberos")
                    )
                )
                and (self.connect_to_apache_hive)
            )
            else exclude.add("hive_username")
        )
        include.add("hive_database") if (self.connect_to_apache_hive) else exclude.add("hive_database")
        (
            include.add("hive_service_principal_name")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
                and (self.connect_to_apache_hive)
            )
            else exclude.add("hive_service_principal_name")
        )
        (
            include.add("hive_user_principal_name")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
                and (self.connect_to_apache_hive)
            )
            else exclude.add("hive_user_principal_name")
        )
        (
            include.add("service_principal_name")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
            )
            else exclude.add("service_principal_name")
        )
        (
            include.add("password")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value != "kerberos"
                        )
                        or (self.authentication_method != "kerberos")
                    )
                )
            )
            else exclude.add("password")
        )
        include.add("hive_host") if (self.connect_to_apache_hive) else exclude.add("hive_host")
        (
            include.add("user_principal_name")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
            )
            else exclude.add("user_principal_name")
        )
        (
            include.add("hive_keytab_file")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "kerberos"
                        )
                        or (self.authentication_method == "kerberos")
                    )
                )
                and (self.connect_to_apache_hive)
            )
            else exclude.add("hive_keytab_file")
        )
        (
            include.add("username")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value != "kerberos"
                        )
                        or (self.authentication_method != "kerberos")
                    )
                )
            )
            else exclude.add("username")
        )
        (
            include.add("satellite_connector_id")
            if (((not self.secure_gateway_id)) and (not self.satellite_location_id))
            else exclude.add("satellite_connector_id")
        )
        (
            include.add("satellite_location_id")
            if (((not self.secure_gateway_id)) and (not self.satellite_connector_id))
            else exclude.add("satellite_location_id")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("authentication_method") if (self.hidden_dummy_property1) else exclude.add("authentication_method")
        (
            include.add("hive_user_principal_name")
            if (self.hidden_dummy_property1)
            else exclude.add("hive_user_principal_name")
        )
        (
            include.add("service_principal_name")
            if (self.hidden_dummy_property1)
            else exclude.add("service_principal_name")
        )
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")
        include.add("keytab_file") if (self.hidden_dummy_property1) else exclude.add("keytab_file")
        include.add("user_principal_name") if (self.hidden_dummy_property1) else exclude.add("user_principal_name")
        (
            include.add("hive_login_config_name")
            if (self.hidden_dummy_property1)
            else exclude.add("hive_login_config_name")
        )
        include.add("hive_keytab_file") if (self.hidden_dummy_property1) else exclude.add("hive_keytab_file")
        include.add("file_system_type") if (self.hidden_dummy_property1) else exclude.add("file_system_type")
        include.add("ssl_certificate_file") if (self.hidden_dummy_property1) else exclude.add("ssl_certificate_file")
        (
            include.add("hive_service_principal_name")
            if (self.hidden_dummy_property1)
            else exclude.add("hive_service_principal_name")
        )

        (
            include.add("satellite_connector_id")
            if (not self.secure_gateway_id) and (not self.satellite_location_id)
            else exclude.add("satellite_connector_id")
        )
        (
            include.add("user_principal_name")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            else exclude.add("user_principal_name")
        )
        (
            include.add("hive_keytab_file")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_keytab_file")
        )
        (
            include.add("password")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value != "kerberos")
                    or (self.authentication_method != "kerberos")
                )
            )
            else exclude.add("password")
        )
        (
            include.add("username")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value != "kerberos")
                    or (self.authentication_method != "kerberos")
                )
            )
            else exclude.add("username")
        )
        include.add("ssl_certificate") if (not self.ssl_certificate_file) else exclude.add("ssl_certificate")
        (
            include.add("hive_username")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value != "kerberos")
                    or (self.authentication_method != "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_username")
        )
        (
            include.add("hive_service_principal_name")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_service_principal_name")
        )
        (
            include.add("satellite_location_id")
            if (not self.secure_gateway_id) and (not self.satellite_connector_id)
            else exclude.add("satellite_location_id")
        )
        (
            include.add("hive_user_principal_name")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_user_principal_name")
        )
        (
            include.add("service_principal_name")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            else exclude.add("service_principal_name")
        )
        (
            include.add("hive_port")
            if (self.connect_to_apache_hive and "true" in str(self.connect_to_apache_hive))
            else exclude.add("hive_port")
        )
        (
            include.add("hive_database")
            if (self.connect_to_apache_hive and "true" in str(self.connect_to_apache_hive))
            else exclude.add("hive_database")
        )
        (
            include.add("hive_password")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value != "kerberos")
                    or (self.authentication_method != "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_password")
        )
        (
            include.add("hive_http_path")
            if (self.connect_to_apache_hive and "true" in str(self.connect_to_apache_hive))
            else exclude.add("hive_http_path")
        )
        (
            include.add("keytab_file")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            else exclude.add("keytab_file")
        )
        (
            include.add("hive_login_config_name")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "kerberos")
                    or (self.authentication_method == "kerberos")
                )
            )
            and (self.connect_to_apache_hive == "true" or self.connect_to_apache_hive)
            else exclude.add("hive_login_config_name")
        )
        include.add("ssl_certificate_file") if (not self.ssl_certificate) else exclude.add("ssl_certificate_file")
        (
            include.add("hive_host")
            if (self.connect_to_apache_hive and "true" in str(self.connect_to_apache_hive))
            else exclude.add("hive_host")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""This module contains PythonGenerator class."""

from ibm_watsonx_data_integration.codegen.processors.streaming_processor import StreamingProcessor
from ibm_watsonx_data_integration.codegen.sources import source_factory
from ibm_watsonx_data_integration.common.constants import (
    PROD_BASE_API_URL,
    PROD_BASE_URL,
)
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.codegen.code import Coder
    from ibm_watsonx_data_integration.common.auth import IAMAuthenticator
    from ibm_watsonx_data_integration.cpd_models.flow_model import Flow


class PythonGenerator:
    """Flow code generation entrypoint."""

    def __init__(
        self,
        source: "Flow",
        destination: str | Path,
        auth: "IAMAuthenticator",  # pragma: allowlist secret
        **kwargs: str | object,
    ) -> None:
        """The __init__ of the PythonGenerator class.

        Args:
            source: Location from ``Flow`` definition will be loaded.
            destination: Location where save generated script.
            auth: Reference to authenticator object.
            kwargs: Additional configuration values.
        """
        self._source = source_factory(source)
        self._destination = Path(destination)
        base_url = kwargs.get("base_url", PROD_BASE_URL)
        base_api_url = kwargs.get("base_api_url", PROD_BASE_API_URL)
        self._processor: Coder = StreamingProcessor(
            source_data=self._source.to_json(), auth=auth, base_url=base_url, base_api_url=base_api_url
        )

    def save(self) -> Path:
        """Run code generation then save it to destination.

        Returns:
            Path to the location where the generated script was saved.
        """
        code = self._processor.to_code()
        return code.save(self._destination)



# -------------------------------------------------------------------
# This file was auto-generated. Do not edit directly.
# Generated by: StageGeneratorV2
# -------------------------------------------------------------------

LABEL_MAPPINGS: dict[str, str] = {
    "address_verification": "Address Verification",
    "aggregator": "Aggregator",
    "amazon_postgresql": "Amazon RDS for PostgreSQL",
    "amazon_redshift": "Amazon Redshift",
    "amazonrds_oracle": "Amazon RDS for Oracle",
    "amazons3": "Amazon S3",
    "apache_hbase": "Apache HBase",
    "apache_hive": "Apache Hive",
    "apache_kafka": "Apache Kafka",
    "azure_blob_storage": "Microsoft Azure Blob Storage",
    "azure_cosmos": "Microsoft Azure Cosmos DB",
    "azure_databricks": "Microsoft Azure Databricks",
    "azure_file_storage": "Microsoft Azure File Storage",
    "azure_postgresql": "Azure PostgreSQL",
    "azuredatalake": "Microsoft Azure Data Lake Storage",
    "azuresql": "Microsoft Azure SQL Database",
    "azuresynapse": "Microsoft Azure Synapse Analytics",
    "bigquery": "Google BigQuery",
    "bigsql": "IBM Db2 Big SQL",
    "bloom_filter": "Bloom Filter",
    "box": "Box",
    "cassandra": "Apache Cassandra",
    "cassandra_datastage": "Apache Cassandra for DataStage",
    "change_apply": "Change Apply",
    "change_capture": "Change Capture",
    "checksum": "Checksum",
    "cloud_object_storage": "IBM Cloud Object Storage",
    "cognos_analytics": "IBM Cognos Analytics",
    "column_export": "Column Export",
    "column_generator": "Column Generator",
    "column_import": "Column Import",
    "combine_records": "Combine Records",
    "compare": "Compare",
    "complex_flat_file": "Complex Flat File",
    "compress": "Compress",
    "copy": "Copy",
    "data_rules": "Data Rules",
    "dataset": "Data set",
    "datastax": "DataStax Enterprise",
    "db2": "IBM Db2",
    "db2cloud": "IBM Db2 on Cloud",
    "db2fordatastage": "IBM Db2 for DataStage",
    "db2iseries": "IBM Db2 for i",
    "db2warehouse": "IBM Db2 Warehouse",
    "db2zos": "IBM Db2 for z/OS",
    "decode": "Decode",
    "denodo": "Denodo",
    "derby": "Apache Derby",
    "difference": "Difference",
    "dremio": "Dremio",
    "dropbox": "Dropbox",
    "dv": "IBM Data Virtualization",
    "dvm": "IBM Data Virtualization Manager for z/OS",
    "elasticsearch": "Elasticsearch",
    "encode": "Encode",
    "exasol": "Exasol",
    "expand": "Expand",
    "external_filter": "External Filter",
    "external_source": "External Source",
    "external_target": "External Target",
    "fileset": "File set",
    "filter": "Filter",
    "ftp": "FTP",
    "funnel": "Funnel",
    "generic": "Generic",
    "generics3": "Generic S3",
    "google_cloud_storage": "Google Cloud Storage",
    "google_looker": "Google Looker",
    "google_pub_sub": "Google Cloud Pub/Sub",
    "greenplum": "Greenplum",
    "hdfs_apache": "Apache HDFS",
    "head": "Head",
    "hierarchical_data": "Hierarchical Data",
    "http": "HTTP",
    "ibm_mq": "IBM MQ",
    "impala": "Apache Impala",
    "informix": "IBM Informix",
    "investigate": "Investigate",
    "java_integration": "Java Integration",
    "jdbc": "Generic JDBC",
    "join": "Join",
    "lookup": "Lookup",
    "lookupfileset": "Lookup file set",
    "make_subrecord": "Make Subrecord",
    "make_vector": "Make Vector",
    "mariadb": "MariaDB",
    "match360": "IBM Match 360",
    "match_frequency": "Match Frequency",
    "merge": "Merge",
    "minio": "MinIO",
    "modify": "Modify",
    "mongodb": "MongoDB",
    "mongodb_ibmcloud": "IBM Cloud Databases for MongoDB",
    "mysql": "MySQL",
    "mysql_amazon": "Amazon RDS for MySQL",
    "mysql_compose": "IBM Cloud Databases for MySQL",
    "netezza": "IBM Netezza Performance Server",
    "netezza_optimized": "IBM Netezza Performance Server for DataStage",
    "odbc": "ODBC",
    "one_source_match": "One-source Match",
    "oracle": "Oracle",
    "oracle_datastage": "Oracle Database for DataStage",
    "peek": "Peek",
    "pivot": "Pivot Enterprise",
    "planning_analytics": "IBM Planning Analytics",
    "postgresql": "PostgreSQL",
    "postgresql_ibmcloud": "IBM Cloud Databases for PostgreSQL",
    "presto": "Presto",
    "promote_subrecord": "Promote Subrecord",
    "remove_duplicates": "Remove Duplicates",
    "rest": "Rest",
    "row_generator": "Row Generator",
    "salesforce": "Salesforce.com",
    "salesforceapi": "Salesforce API for DataStage",
    "sample": "Sample",
    "sapbapi": "SAP BAPI",
    "sapbulkextract": "SAP Bulk Extract",
    "sapdeltaextract": "SAP Delta Extract",
    "saphana": "SAP HANA",
    "sapidoc": "SAP IDoc",
    "sapiq": "SAP IQ",
    "sapodata": "SAP OData",
    "sequentialfile": "Sequential file",
    "singlestore": "SingleStoreDB",
    "slowly_changing_dimension": "Slowly Changing Dimension",
    "snowflake": "Snowflake",
    "sort": "Sort",
    "split_subrecord": "Split Subrecord",
    "split_vector": "Split Vector",
    "sqlserver": "Microsoft SQL Server",
    "standardize": "Standardize",
    "storage_volume": "Storage volume",
    "surrogate_key_generator": "Surrogate Key Generator",
    "survive": "Survive",
    "switch": "Switch",
    "sybase": "SAP ASE",
    "tableau": "Tableau",
    "tail": "Tail",
    "teradata": "Teradata",
    "teradata_datastage": "Teradata database for DataStage",
    "transformer": "Transformer",
    "two_source_match": "Two-source Match",
    "vertica": "Vertica",
    "watsonx_data": "IBM watsonx.data Presto",
    "wave_generator": "Wave Generator",
    "web_service": "Web Service",
    "write_range_map": "Write Range Map",
    "xml_input": "XML Input",
    "xml_output": "XML Output",
}



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import APACHE_HBASE_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class FilePathMode(Enum):
    content = "content"
    path = "path"


class ApacheHbaseConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "4f4f7244-9459-3f91-a149-f49f4cc59c54"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    cas_lite_service_authorization_header: str | None = Field(None, alias="cas_lite_auth_header")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    core_site_xml: str = Field(None, alias="core-site")
    core_site_xml_path: str = Field(None, alias="core-site_path")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    input_method_for_xml_files: APACHE_HBASE_CONNECTION.FilePathMode | None = Field(
        APACHE_HBASE_CONNECTION.FilePathMode.content, alias="file_path_mode"
    )
    hadoop_identity: str | None = Field("", alias="hadoop_identity")
    hbase_site_xml: str = Field(None, alias="hbase-site")
    hbase_site_xml_path: str = Field(None, alias="hbase-site_path")
    hbase_identity: str | None = Field("", alias="hbase_identity")
    simple_authentication_sasl_username: str = Field(None, alias="simple_auth_username")
    use_cas_lite_service: bool | None = Field(True, alias="use_cas_lite")
    use_kerberos_authentication: bool = Field(False, alias="use_kerberos")
    krb5_conf_location: str = Field(None, alias="use_kerberos.krb5conf")
    password: str = Field("password", alias="use_kerberos.password")
    user_principal_name: str = Field("user@REALM", alias="use_kerberos.principal")
    ticket_cache_location: str | None = Field(None, alias="use_kerberos.ticket_cache_location")
    use_keytab: bool = Field(False, alias="use_kerberos.use_keytab")
    keytab: str = Field("", alias="use_kerberos.use_keytab.keytab")
    use_ticket_cache: bool = Field(False, alias="use_kerberos.use_ticket_cache")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        include.add("keytab") if (((self.use_kerberos_authentication)) and (self.use_keytab)) else exclude.add("keytab")
        include.add("krb5_conf_location") if (self.use_kerberos_authentication) else exclude.add("krb5_conf_location")
        (
            include.add("core_site_xml_path")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "path"
                    )
                    or (self.input_method_for_xml_files == "path")
                )
            )
            else exclude.add("core_site_xml_path")
        )
        (
            include.add("simple_authentication_sasl_username")
            if (
                ((not self.defer_credentials))
                and (((not self.use_kerberos_authentication)) or (not self.use_kerberos_authentication))
            )
            else exclude.add("simple_authentication_sasl_username")
        )
        (
            include.add("use_ticket_cache")
            if (((self.use_kerberos_authentication)) and (not self.use_keytab))
            else exclude.add("use_ticket_cache")
        )
        include.add("use_keytab") if (self.use_kerberos_authentication) else exclude.add("use_keytab")
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        (
            include.add("hbase_site_xml")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "content"
                    )
                    or (self.input_method_for_xml_files == "content")
                )
            )
            else exclude.add("hbase_site_xml")
        )
        include.add("user_principal_name") if (self.use_kerberos_authentication) else exclude.add("user_principal_name")
        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("ticket_cache_location")
            if (((self.use_kerberos_authentication)) and (not self.use_keytab) and (self.use_ticket_cache))
            else exclude.add("ticket_cache_location")
        )
        (
            include.add("password")
            if ((((self.use_kerberos_authentication))) and (not self.use_keytab))
            else exclude.add("password")
        )
        (
            include.add("core_site_xml")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "content"
                    )
                    or (self.input_method_for_xml_files == "content")
                )
            )
            else exclude.add("core_site_xml")
        )
        (
            include.add("hbase_site_xml_path")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "path"
                    )
                    or (self.input_method_for_xml_files == "path")
                )
            )
            else exclude.add("hbase_site_xml_path")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )

        (
            include.add("use_keytab")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            else exclude.add("use_keytab")
        )
        (
            include.add("ticket_cache_location")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            and (self.use_keytab == "false" or not self.use_keytab)
            and (self.use_ticket_cache == "true" or self.use_ticket_cache)
            else exclude.add("ticket_cache_location")
        )
        (
            include.add("hbase_site_xml")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "content"
                    )
                    or (self.input_method_for_xml_files == "content")
                )
            )
            else exclude.add("hbase_site_xml")
        )
        (
            include.add("hbase_site_xml_path")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "path"
                    )
                    or (self.input_method_for_xml_files == "path")
                )
            )
            else exclude.add("hbase_site_xml_path")
        )
        (
            include.add("krb5_conf_location")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            else exclude.add("krb5_conf_location")
        )
        (
            include.add("keytab")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            and (self.use_keytab == "true" or self.use_keytab)
            else exclude.add("keytab")
        )
        (
            include.add("use_cas_lite_service")
            if (self.cas_lite_service_authorization_header)
            else exclude.add("use_cas_lite_service")
        )
        (
            include.add("user_principal_name")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            else exclude.add("user_principal_name")
        )
        (
            include.add("use_ticket_cache")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            and (self.use_keytab == "false" or not self.use_keytab)
            else exclude.add("use_ticket_cache")
        )
        (
            include.add("core_site_xml")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "content"
                    )
                    or (self.input_method_for_xml_files == "content")
                )
            )
            else exclude.add("core_site_xml")
        )
        (
            include.add("simple_authentication_sasl_username")
            if (not self.defer_credentials)
            and (
                (not self.use_kerberos_authentication)
                or (self.use_kerberos_authentication == "false" or not self.use_kerberos_authentication)
            )
            else exclude.add("simple_authentication_sasl_username")
        )
        (
            include.add("cas_lite_service_authorization_header")
            if (self.use_cas_lite_service == "true" or self.use_cas_lite_service)
            else exclude.add("cas_lite_service_authorization_header")
        )
        (
            include.add("core_site_xml_path")
            if (
                self.input_method_for_xml_files
                and (
                    (
                        hasattr(self.input_method_for_xml_files, "value")
                        and self.input_method_for_xml_files.value == "path"
                    )
                    or (self.input_method_for_xml_files == "path")
                )
            )
            else exclude.add("core_site_xml_path")
        )
        (
            include.add("password")
            if (self.use_kerberos_authentication == "true" or self.use_kerberos_authentication)
            and (self.use_keytab == "false" or not self.use_keytab)
            else exclude.add("password")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



from ibm_watsonx_data_integration.services.datastage.models.enums import TERADATA_CONNECTION
from os import environ
from tests.integration.datastage.util import AbstractFlowRunTest
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestTeradataConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "teradata_peek"

    def get_fc(self, project: Project):
        fc = project.create_flow(name=self.get_flow_name(), environment=None, description="", flow_type="datastage")

        # teradataconn = TeradataConn(
        #     name="teradataconn",
        #     authentication_method=TERADATA_CONNECTION.AuthenticationMethod.td2,
        #     database="tm_ds_db_1",
        #     hostname_or_ip_address=environ["TERADATA_HOST"],
        #     password=environ["TERADATA_PASSWORD"],
        #     port_is_ssl_enabled=False,
        #     enable_trusted_session=False,
        #     username=environ["TERADATA_USERNAME"],
        # )
        # Stages
        print(environ["TERADATA_HOST"])
        _fastpathtest_1_ = fc.add_stage("Teradata", "fastpathTest_1")
        _fastpathtest_1_.configuration.schema_name = "tm_ds_db_1"
        _fastpathtest_1_.configuration.table_name = "fastpathTest"

        _fastpathtest_1_.configuration.connection.authentication_method = TERADATA_CONNECTION.AuthenticationMethod.td2
        _fastpathtest_1_.configuration.connection.database = "tm_ds_db_1"
        _fastpathtest_1_.configuration.connection.hostname_or_ip_address = environ["TERADATA_HOST"]
        _fastpathtest_1_.configuration.connection.password = environ["TERADATA_PASSWORD"]
        _fastpathtest_1_.configuration.connection.port_is_ssl_enabled = False
        _fastpathtest_1_.configuration.connection.enable_trusted_session = False
        _fastpathtest_1_.configuration.connection.username = environ["TERADATA_USERNAME"]
        # _fastpathtest_1_.configuration.connection = teradataconn

        _peek_1_ = fc.add_stage("Peek", "Peek_1")
        _peek_1_.configuration.runtime_column_propagation = 0

        # Graph
        fastpathtest_1_schema = _fastpathtest_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        fastpathtest_1_schema.add_field("INTEGER", "a")
        fastpathtest_1_schema.add_field("CHAR", "b").length(1)
        fastpathtest_1_schema.add_field("CHAR", "c").length(1)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.enums import REST
from pydantic import Field
from typing import ClassVar


class rest(BaseStage):

    op_name: ClassVar[str] = "PxRest"
    node_type: ClassVar[str] = "execution_node"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/REST.svg"
    label: ClassVar[str] = "Rest"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    combinability: REST.Combinability | None = Field(REST.Combinability.auto, alias="combinability")
    execmode: REST.Execmode | None = Field(REST.Execmode.default_seq, alias="execmode")
    input_count: int | None = Field(0, alias="input_count")
    is_reject_link: bool | None = Field(False, alias="is_reject_link")
    link_name: str | None = Field(" ", alias="link_name")
    log_level: REST.LogLevel | None = Field(REST.LogLevel.warning, alias="log_level")
    log_reject_error: bool | None = Field(False, alias="log_reject_error")
    output_count: int | None = Field(0, alias="output_count")
    parameters: str | None = Field(" ", alias="parameters")
    part_coll: str | None = Field("part_type", alias="part_coll")
    preserve: REST.Preserve | None = Field(REST.Preserve.default_propagate, alias="preserve")
    reject_message_column: REST.RejectMessageColumn | None = Field(
        REST.RejectMessageColumn.custom, alias="reject_message_column"
    )
    requests: list = Field([], alias="requests")
    runtime_column_propagation: int | None = Field(1, alias="runtime_column_propagation")
    variables: dict = Field([], alias="variables")

    def validate(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        return include, exclude

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {"link_name", "part_coll", "runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 0, "max": -1}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {"is_reject_link", "reject_message_column"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 0, "max": -1}

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate()
        props = {
            "combinability",
            "execmode",
            "input_count",
            "log_level",
            "log_reject_error",
            "output_count",
            "parameters",
            "preserve",
            "requests",
            "variables",
        }
        required = {"requests", "variables"}
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "maxRejectOutputs": 0,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing the StreamsetsFlow API client."""

import requests
from ibm_watsonx_data_integration.common.utils import wait_and_retry_on_http_error
from ibm_watsonx_data_integration.cpd_api.base import BaseAPIClient
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator

FLOWSTORE_API_VERSION = 1


class StreamsetsFlowApiClient(BaseAPIClient):
    """The API client of the Project."""

    def __init__(
        self,
        auth: "BaseAuthenticator",  # pragma: allowlist secret
        base_url: str = "https://api.dataplatform.cloud.ibm.com",
    ) -> None:
        """The __init__ of the ProjectApiClient.

        Args:
            auth: The Authentication object.
            base_url: The Streamsets Flow URL.
        """
        super().__init__(auth=auth, base_url=base_url)
        self.url_path = f"sset/streamsets_flows/v{FLOWSTORE_API_VERSION}"
        self.url_path_connections = f"sset/streamsets_flows/v{FLOWSTORE_API_VERSION}/connections"

    @wait_and_retry_on_http_error(timeout_sec=10)
    def get_streamsets_flows(self, params: dict) -> requests.Response:
        """Get all Projects.

        Args:
           params: The Query params.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows"
        response = self.get(url=url, params=params)
        return response

    def get_streamsets_flow_by_id(self, params: dict, flow_id: str) -> requests.Response:
        """Get all Projects.

        Args:
            params: The Query params.
            flow_id: The Flow ID.

        Returns:
           A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows/{quote(flow_id, safe='')}"
        response = self.get(url=url, params=params)
        return response

    def delete_streamsets_flow(self, params: dict, flow_id: str) -> requests.Response:
        """Delete a Streamsets Flow.

        Args:
            params: The Query params.
            flow_id: The Flow ID.

        Returns:
           A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows/{quote(flow_id, safe='')}"
        response = self.delete(url=url, params=params)
        return response

    def update_streamsets_flow(self, params: dict, flow_id: str, data: str) -> requests.Response:
        """Update a Streamsets Flow.

        Args:
            params: The Query params.
            flow_id: The Flow ID.
            data: The payload data.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows/{quote(flow_id, safe='')}"
        response = self.put(url=url, data=data, params=params)
        return response

    @wait_and_retry_on_http_error(timeout_sec=4)
    def create_streamsets_flow(self, params: dict, data: str) -> requests.Response:
        """Create a Streamsets Flow.

        Args:
            params: The Query params.
            data: The payload data.

        Returns:
           A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows"
        response = self.post(url=url, data=data, params=params)
        return response

    def duplicate_streamsets_flow(self, params: dict, flow_id: str, data: str) -> requests.Response:
        """Duplicate a Streamsets Flow.

        Args:
            params: The Query params.
            flow_id: The Flow ID.
            data: The payload data.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/streamsets_flows/{quote(flow_id, safe='')}/duplicate"
        response = self.post(url=url, data=data, params=params)
        return response

    def get_streamsets_connection(self, connection_id: str, params: dict[str, Any]) -> requests.Response:
        """Get StreamSets connection.

        Args:
            connection_id: Connection id.
            params: REST Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_connections}/{quote(connection_id, safe='')}"
        return self.get(url=url, params=params)

    def get_streamsets_connections(self, params: dict[str, Any]) -> requests.Response:
        """List defined StreamSets connections.

        Args:
            params: REST Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_connections}"
        return self.get(url=url, params=params)



"""This module converts a flow DAG (Directed Acyclic Graph) into structured Pydantic models for
a JSON representation of the flow ("flow JSON").

Algorithm overview:
- The `FlowExtractor` initializes the conversion process, managing multiple `PipelineExtractor` instances.
- Each `PipelineExtractor` processes a `DAG`, creating node, link, and port representations.
- Nodes in the DAG are mapped to `AbstractNodeExtractor` subclasses (`ExecutionNodeExtractor`,
  `BindingNodeExtractor`, etc.) based on their type.
- Edges in the DAG are converted into `LinkExtractor` objects, while their connections are handled by
  `InputPortExtractor` and `OutputPortExtractor`.
- Supernodes, representing subflows, are processed via `SuperNodeExtractor`.

Classes:
- `FlowExtractor`: Manages pipeline extraction from the DAG.
- `PipelineExtractor`: Represents a pipeline, mapping nodes and links.
- `AbstractNodeExtractor` (and subclasses): Converts DAG nodes into model representations.
- `LinkExtractor`: Handles links between nodes.
- `InputPortExtractor` / `OutputPortExtractor`: Defines node ports that links may connect to.

The final output is a `model.Flow` object, encapsulating the DataStage flow in a format that can be directly dumped as
a flow JSON importable into the canvas UI or run as a job.
"""

import json
import ibm_watsonx_data_integration.services.datastage.models.flow_json_model as model
from abc import ABC, abstractmethod
import pydantic
from ibm_watsonx_data_integration.services.datastage.models.acp import ACP
from ibm_watsonx_data_integration.services.datastage.models.flow.dag import (
    DAG,
    Link,
    StageNode,
    Node,
    StageNode,
)
from ibm_watsonx_data_integration.services.datastage.models.schema.schema import Schema

from typing import Any
from uuid import uuid4

from ibm_watsonx_data_integration.services.datastage.models.stage_models.address_verification import address_verification
from ibm_watsonx_data_integration.services.datastage.models.stage_models.aggregator import aggregator
from ibm_watsonx_data_integration.services.datastage.models.stage_models.bloom_filter import bloom_filter
from ibm_watsonx_data_integration.services.datastage.models.stage_models.change_apply import change_apply
from ibm_watsonx_data_integration.services.datastage.models.stage_models.change_capture import change_capture
from ibm_watsonx_data_integration.services.datastage.models.stage_models.checksum import checksum
from ibm_watsonx_data_integration.services.datastage.models.stage_models.column_export import column_export
from ibm_watsonx_data_integration.services.datastage.models.stage_models.column_generator import column_generator
from ibm_watsonx_data_integration.services.datastage.models.stage_models.column_import import column_import
from ibm_watsonx_data_integration.services.datastage.models.stage_models.compare import compare
from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_flat_file import complex_flat_file
from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_stages.complex_flat_file import Record, RecordID
from ibm_watsonx_data_integration.services.datastage.models.stage_models.difference import difference
from ibm_watsonx_data_integration.services.datastage.models.stage_models.funnel import funnel
from ibm_watsonx_data_integration.services.datastage.models.stage_models.generic import generic
from ibm_watsonx_data_integration.services.datastage.models.stage_models.join import join
from ibm_watsonx_data_integration.services.datastage.models.stage_models.lookup import lookup
from ibm_watsonx_data_integration.services.datastage.models.stage_models.match_frequency import match_frequency
from ibm_watsonx_data_integration.services.datastage.models.stage_models.peek import peek
from ibm_watsonx_data_integration.services.datastage.models.stage_models.pivot import pivot
from ibm_watsonx_data_integration.services.datastage.models.stage_models.sort import sort
from ibm_watsonx_data_integration.services.datastage.models.stage_models.standardize import standardize
from ibm_watsonx_data_integration.services.datastage.models.stage_models.transformer import transformer

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ibm_watsonx_data_integration.services.datastage.models.flow import DataStageFlow

import logging
logger = logging.getLogger(__name__)


def generate_uuid():
    """Generate a unique identifier as a string."""
    return uuid4().__str__()


def satisfies_min_cardinality(cardinality: int, min_cardinality: int):
    """Returns True if the cardinality is greater than or equal to the minimum cardinality. If the minimum cardinality
    is -1, it is ignored.
    """
    return min_cardinality == -1 or cardinality >= min_cardinality


def satisfies_max_cardinality(cardinality: int, max_cardinality: int):
    """Returns True if the cardinality is less than or equal to the maximum cardinality. If the maximum cardinality
    is -1, it is ignored.
    """
    return max_cardinality == -1 or cardinality <= max_cardinality


class AbstractIdent(ABC):
    """Abstract class providing a unique identifier for each instance."""

    def __init__(self):
        self.__id = generate_uuid()

    @property
    def id(self):
        return self.__id


def verify_node_cardinality(node: Node, dag: DAG):
    assert dag.is_metadata_computed

    primary_in = node.metadata.get("primary_inputs", 0)
    primary_out = node.metadata.get("primary_outputs", 0)
    reject_out = node.metadata.get("reject_outputs", 0)
    reference_in = node.metadata.get("reference_inputs", 0)

    # Get expected cardinality
    min_primary_in = node._get_min_primary_inputs()
    max_primary_in = node._get_max_primary_inputs()
    min_primary_out = node._get_min_primary_outputs()
    max_primary_out = node._get_max_primary_outputs()
    min_reject_out = node._get_min_reject_outputs()
    max_reject_out = node._get_max_reject_outputs()
    min_reference_in = node._get_min_reference_inputs()
    max_reference_in = node._get_max_reference_inputs()

    # Check primary input cardinality
    if not satisfies_min_cardinality(primary_in, min_cardinality=min_primary_in):
        logger.warning(
            f"\n{node.__class__.__name__} {node} has insufficient primary inputs: {primary_in}, min: {min_primary_in}",
        )
    if not satisfies_max_cardinality(primary_in, max_cardinality=max_primary_in):
        raise ValueError(
            f"\n{node.__class__.__name__} {node} has too many primary inputs: {primary_in}, max: {max_primary_in}"
        )

    # Check primary output cardinality
    if not satisfies_min_cardinality(primary_out, min_cardinality=min_primary_out):
        logger.warning(
            f"\n{node.__class__.__name__} {node} has insufficient primary outputs: {primary_out}, min: {min_primary_out}",
        )
    if not satisfies_max_cardinality(primary_out, max_cardinality=max_primary_out):
        raise ValueError(
            f"\n{node.__class__.__name__} {node} has too many primary outputs: {primary_out}, max: {max_primary_out}"
        )

    # Check reject output cardinality
    if not satisfies_min_cardinality(reject_out, min_cardinality=min_reject_out):
        logger.warning(
            f"\n{node.__class__.__name__} {node} has insufficient reject outputs: {reject_out}, min: {min_reject_out}",
        )
    if not satisfies_max_cardinality(reject_out, max_cardinality=max_reject_out):
        raise ValueError(
            f"\n{node.__class__.__name__} {node} has too many reject outputs: {reject_out}, max: {max_reject_out}"
        )

    # Check reference input cardinality
    if not satisfies_min_cardinality(reference_in, min_cardinality=min_reference_in):
        logger.warning(
            f"\n{node.__class__.__name__} {node} has insufficient reference inputs: {reference_in}, min: {min_reference_in}",
        )
    if not satisfies_max_cardinality(reference_in, max_cardinality=max_reference_in):
        raise ValueError(
            f"\n{node.__class__.__name__} {node} has too many reference inputs: {reference_in}, max: {max_reference_in}"
        )

    if isinstance(node, StageNode):
        node.configuration.input_count = primary_in + reference_in
        node.configuration.output_count = primary_out + reject_out


class FlowExtractor(AbstractIdent):
    """Extracts a flow model from a flow DAG. This is the master class that accepts a flow DAG and represents its
    entirety as a flow model.

    To use: instantiate FlowExtractor with your DAG, then call extract(), and then call serialize().
    """

    def __init__(self, fc: "DataStageFlow"):
        super().__init__()
        self.fc = fc
        self.top_level_dag = fc._dag
        self.all_nodes: dict[Node, AbstractNodeExtractor] = {}
        self.schemas: list[SchemaExtractor] = []
        self.pipelines: dict[DAG, PipelineExtractor] = {}

    def get_or_create_node_extractor(self, pipeline: "PipelineExtractor", node: Node):
        """Creates a node extractor based on its type."""
        if ext := self.all_nodes.get(node):
            return ext

        if isinstance(node, StageNode):
            if hasattr(node.configuration, "connection"):
                ext = ConnectorNodeExtractor(pipeline, node)
            else:
                match node._get_node_type():
                    case "binding":
                        ext = BindingNodeExtractor(pipeline, node)
                    case "execution_node":
                        ext = ExecutionNodeExtractor(pipeline, node)

        if ext:
            self.all_nodes[node] = ext
            return ext
        else:
            raise NotImplementedError(f"Unsupported node type: {type(node)}")

    def extract(self):
        """Extracts complete pipelines from the DAG."""
        # Add the top level DAG as the primary pipeline
        self.pipelines[self.top_level_dag] = PipelineExtractor(self, self.top_level_dag)

        self.pipelines[self.top_level_dag].extract()

        # Verify the cardinality of every node in the flow
        for node in self.all_nodes.keys():
            verify_node_cardinality(node, self.top_level_dag)

        # Apply global RCP value to all nodes
        if self.fc.rcp:
            for node in self.all_nodes.keys():
                if isinstance(node, StageNode):
                    node.configuration.runtime_column_propagation = True

    def serialize(self):
        """Serializes the flow into a model.Flow object."""
        if not self.pipelines:
            raise ValueError("No pipelines found. Did you call extract()?")

        pipelines: list[model.Pipeline] = []
        for pipe in self.pipelines.values():
            pipelines.append(pipe.serialize())

        primary_pipeline = self.pipelines[self.top_level_dag]

        app_data = model.AppData()
        app_data.additionalProperties = {"globalAcp": self.fc.acp}

        return model.Flow(
            id=self.id,
            doc_type="pipeline",
            primary_pipeline=primary_pipeline.id,
            pipelines=pipelines,
            schemas=[sch.serialize() for sch in self.schemas],
            app_data=app_data,
        )

class PipelineExtractor(AbstractIdent):
    """Extracts a complete pipeline model, including all of its nodes, links, and ports, from a DAG."""

    def __init__(
        self,
        parent_flow: FlowExtractor,
        dag: DAG,
    ):
        super().__init__()
        self.parent_flow = parent_flow
        self.dag = dag
        self.used_link_names = set()
        self.links: dict[Link, LinkExtractor] = {}
        self.input_ports: dict[Link, InputPortExtractor] = {}
        self.output_ports: dict[Link, OutputPortExtractor] = {}
        self.schemas: dict[str, model.RecordSchema] = {}


    def extract(self):
        """Extracts nodes, links, and ports from the pipeline."""
        # Reset link counter to 1
        LinkExtractor.reset_counter()

        if not self.dag.is_metadata_computed:
            self.dag.compute_metadata()

        for link in self.dag.links_stable():
            # Get or create a link extractor while enforcing link name uniqueness
            if link not in self.links:
                link_ext = self.links.setdefault(link, LinkExtractor(self, link))
                self.used_link_names.add(link_ext.get_link_name())
            else:
                link_ext = self.links[link]

            src_node = self.parent_flow.get_or_create_node_extractor(self, link.src)
            dest_node = self.parent_flow.get_or_create_node_extractor(self, link.dest)

            schema: model.RecordSchema | None = None

            # Check if link has a schema attached. If so, use it. Otherwise, fallback to schema attached to src node.
            if link.schema:
                schema = link.schema.configuration
            elif len(src_node.inputs) > 0 or (
                hasattr(src_node.node, "configuration") and src_node.node.configuration.op_name == "PxCFF"
            ):
                input_nodes = src_node.inputs[:]
                for input_node in input_nodes:
                    if not input_node.schema:
                        input_node.schema = SchemaExtractor(parent_flow=self.parent_flow, schema=Schema())

                acp = ACP(
                    link=link,
                    src_input_nodes=input_nodes,
                    src_ext=src_node,
                    dest_ext=dest_node,
                )

                schema = acp.compute_schema()
                #link.schema = schema

            # Create a schema extractor if applicable
            schema_ext: SchemaExtractor | None = None
            if schema:
                schema_ext = SchemaExtractor(self.parent_flow, schema)
                self.parent_flow.schemas.append(schema_ext)

            if isinstance(link.src, StageNode):
                src_output_port = self.output_ports.setdefault(
                    link,
                    OutputPortExtractor(self, link_ext, src_node, schema_ext),
                )

            dest_out_port = None
            if isinstance(link.dest, StageNode):
                dest_input_port = self.input_ports.setdefault(
                    link,
                    InputPortExtractor(self, link_ext, dest_node, schema_ext),
                )

            if link.dest.metadata["out_degree"] == 0:
                dest_out_port = PlainOutputPortExtractor(dest_node)

            dest_node.inputs.append(dest_input_port)
            if dest_out_port:
                if link.dest.metadata["out_degree"] == 0: #not self.is_subflow and :
                    if not len(dest_node.outputs):
                        dest_node.outputs.append(dest_out_port)
                else:
                    dest_node.outputs.append(dest_out_port)
            src_node.outputs.append(src_output_port)

        for node in self.dag.adj:
            if not self.dag.adj[node]:
                self.parent_flow.get_or_create_node_extractor(self, node)

    def serialize(self):
        """Serializes the pipeline into a model.Pipeline object."""
        #comment_models = [comm.serialize() for comm in self.comments.values()]
        node_models = []
        for node in self.dag.nodes():
            #if not isinstance(node, CommentNode):
            mod = self.parent_flow.all_nodes[node].serialize() if node in self.parent_flow.all_nodes else None
            if mod:
                node_models.append(mod)

        return model.Pipeline(
            id=self.id,
            name=None,
            description="",
            runtime_ref="pxOsh",
            app_data=model.PipelineAppData(ui_data=model.PipelineUI()),
            nodes=node_models,
        )

class SchemaExtractor(AbstractIdent):
    def __init__(self, parent_flow: FlowExtractor, schema: model.RecordSchema):
        super().__init__()
        self.parent_flow = parent_flow
        self.schema = schema

    def serialize(self) -> model.RecordSchema:
        self.schema.id = self.id
        return model.RecordSchema(
            id=self.id,
            name=self.schema.name,
            json_schema=self.schema.json_schema,
            type=self.schema.type,
            fields=self.schema.fields,
            struct_types=self.schema.struct_types,
        )


class LinkExtractor(AbstractIdent):
    """Extracts a link between nodes in a pipeline."""

    # Counter keeps track of the global number of instantiated links. It is used to label links with unique numbers.
    counter = 1

    def __init__(self, parent_pipe: PipelineExtractor, link: Link):
        super().__init__()
        self.parent_pipe = parent_pipe
        self.link = link
        self.num = LinkExtractor.counter
        LinkExtractor.counter += 1

    @classmethod
    def reset_counter(cls):
        cls.counter = 1

    def get_link_name(self):
        if self.link.name:
            return self.link.name
        else:
            return f"Link_{self.num}"

    def serialize(self):
        """Serializes the link into a model.NodeLink object."""
        app_data = model.LinkAppData(
            ui_data=model.NodeLinkUI(class_name="", decorations=[]),
        )

        ds_app_data = {}
        if self.link.maps_to_link:
            ds_app_data["partner_link_in"] = self.link.maps_to_link
        if self.link.maps_from_link:
            ds_app_data["partner_link_out"] = self.link.maps_from_link
        app_data.datastage = ds_app_data

        return model.NodeLink(
            id=self.id,
            link_name=self.get_link_name(),
            type_attr=self.link.type,
            node_id_ref=self.parent_pipe.parent_flow.all_nodes[self.link.src].id,
            port_id_ref=self.parent_pipe.output_ports[self.link].id,
            app_data=app_data,
        )


class AbstractNodeExtractor(AbstractIdent):
    """Abstract class representing a node extractor in a pipeline."""

    def __init__(self, parent_pipe: PipelineExtractor, stage_node: StageNode):
        super().__init__()
        self.parent_pipe = parent_pipe
        self.inputs: list[InputPortExtractor] = []
        self.outputs: list[OutputPortExtractor] = []
        self.stage_node: StageNode = stage_node
        self.node: Node = stage_node

    @abstractmethod
    def get_input_port_params(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_output_port_params(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def serialize(self) -> pydantic.BaseModel:
        """Abstract method to serialize the node extractor."""
        pass

    def _serialize_ui_data(self, node: Node) -> model.NodeUI:
        if not isinstance(node, StageNode):
            raise TypeError()
        ui_data = model.NodeUI(
            x_pos=node.metadata.get("x"),
            y_pos=node.metadata.get("y"),
            image=node._get_image(),
            label=node._get_node_label()
            # or self.parent_pipe.ui_gen.lg._get_node_label(node._get_op_name()),
        )

        if width := node.metadata.get("width"):
            ui_data.resize_width = width
            ui_data.is_resized = True
        if height := node.metadata.get("height"):
            ui_data.resize_height = height
            ui_data.is_resized = True

        return ui_data

class ExecutionNodeExtractor(AbstractNodeExtractor):
    """Extracts an execution node within a pipeline."""

    def __init__(self, parent_pipe: PipelineExtractor, stage_node: StageNode):
        super().__init__(parent_pipe, stage_node)
        if not isinstance(stage_node, StageNode):
            raise TypeError()
        self.node: StageNode = stage_node

    def get_input_port_params(self, link=None) -> dict[str, Any]:
        return self.node._get_input_port_params(link)

    def get_output_port_params(self, link=None) -> dict[str, Any]:
        return self.node._get_output_port_params(link)

    def get_output_schema(self) -> model.RecordSchema | None:
        return self.node._get_output_schema()

    def set_output_schema(self, schema: model.RecordSchema):
        self.node._set_output_schema(schema)

    def get_cff_stage_records(self):
        assert isinstance(self.node.configuration, complex_flat_file)
        schemas = []
        stage_records = []
        column_names = []
        for record in self.node.configuration.records:
            for column in record.columns:
                column.mf_update_value = record.name
                if column.name in column_names:
                    logger.warning(
                        f"\nFound duplicate column {column.name} in Complex Flat File stage {self.node.label}",
                    )
                else:
                    column_names.append(column.name)
        for record in self.node.configuration.records:
            assert isinstance(record, Record)
            found_record_id = False
            for record_id in self.node.configuration.records_id:
                assert isinstance(record_id, RecordID)
                if record.name == record_id.record_name:
                    found_record_id = True
                    fields = [column._to_field() for column in record.columns]
                    schema = model.RecordSchema(id="", fields=fields)
                    schema_ext = SchemaExtractor(
                        self.parent_pipe.parent_flow, schema=schema
                    )
                    schemas.append(schema_ext)
                    stage_record = record_id.model_dump(
                        by_alias=True, exclude_none=True
                    )
                    stage_record["schema_ref"] = schema_ext.id
                    stage_records.append(stage_record)
                    break
            if not found_record_id:
                fields = [column._to_field() for column in record.columns]
                schema = model.RecordSchema(id="", fields=fields)
                schema_ext = SchemaExtractor(
                    self.parent_pipe.parent_flow, schema=schema
                )
                schemas.append(schema_ext)
                stage_records.append(
                    {
                        "name": record.name,
                        "record_name": record.name,
                        "schema_ref": schema_ext.id,
                    }
                )
        self.parent_pipe.parent_flow.schemas.extend(schemas)
        if (
            not len(self.node.configuration.records)
            and self.node.configuration.records_id
        ):
            for records_id in self.node.configuration.records_id:
                stage_record = records_id.model_dump(
                    by_alias=True, exclude_none=True, exclude_unset=True
                )
                stage_records.append(stage_record)
        return stage_records

    def get_join_inputs_order(self):
        assert isinstance(self.node.configuration, join)
        if (
            hasattr(self.node.configuration, "inputlink_ordering_list")
            and self.node.configuration.inputlink_ordering_list is not None
        ):
            ordering_list = self.node.configuration.inputlink_ordering_list
            inputs_order = []
            for link in ordering_list:
                if "link_label" in link and link["link_label"] == "Left":
                    link_name = link["link_name"] if "link_name" in link else None
                    if link_name:
                        for input in self.inputs:
                            if input.link.get_link_name() == link_name:
                                inputs_order.append(input.id)
            if len(ordering_list) > 2:
                for i in range(1, len(ordering_list) - 1):
                    for link in ordering_list:
                        if (
                            "link_label" in link
                            and link["link_label"] == f"Intermediate {i}"
                        ):
                            link_name = (
                                link["link_name"] if "link_name" in link else None
                            )
                            if link_name:
                                for input in self.inputs:
                                    if input.link.get_link_name() == link_name:
                                        inputs_order.append(input.id)
            for link in ordering_list:
                if "link_label" in link and link["link_label"] == "Right":
                    link_name = link["link_name"] if "link_name" in link else None
                    if link_name:
                        for input in self.inputs:
                            if input.link.get_link_name() == link_name:
                                inputs_order.append(input.id)
            return "|".join(inputs_order)
        else:
            return None

    def get_lookup_inputs_order(self):
        assert isinstance(self.node.configuration, lookup)
        if (
            hasattr(self.node.configuration, "inputlink_ordering_list")
            and self.node.configuration.inputlink_ordering_list is not None
        ):
            ordering_list = self.node.configuration.inputlink_ordering_list
            inputs_order = []
            for i in range(1, len(ordering_list)):
                for link in ordering_list:
                    if "link_label" in link and link["link_label"] == f"Lookup {i}":
                        link_name = link["link_name"] if "link_name" in link else None
                        if link_name:
                            for input in self.inputs:
                                if input.link.get_link_name() == link_name:
                                    inputs_order.append(input.id)
            for link in ordering_list:
                if "link_label" in link and link["link_label"] == "Primary":
                    link_name = link["link_name"] if "link_name" in link else None
                    if link_name:
                        for input in self.inputs:
                            if input.link.get_link_name() == link_name:
                                inputs_order.append(input.id)
            return "|".join(inputs_order)
        else:
            return None

    def get_lookup_outputs_order(self):
        assert isinstance(self.node.configuration, lookup)
        outputs_order = []
        for output in self.outputs:
            if output.link.link.type == "PRIMARY":
                outputs_order.append(output.id)
        for output in self.outputs:
            if output.link.link.type == "REFERENCE":
                outputs_order.append(output.id)
        for output in self.outputs:
            if output.link.link.type == "REJECT":
                outputs_order.append(output.id)
        return "|".join(outputs_order)

    def get_transformer_outputs_order(self):
        assert isinstance(self.node.configuration, transformer)
        outputs_order = []
        for output in self.outputs:
            if output.link.link.type == "PRIMARY":
                outputs_order.append(output.id)
        for output in self.outputs:
            if output.link.link.type == "REFERENCE":
                outputs_order.append(output.id)
        for output in self.outputs:
            if output.link.link.type == "REJECT":
                outputs_order.append(output.id)
        return "|".join(outputs_order)

    def get_change_capture_inputs_order(self):
        assert isinstance(self.node.configuration, change_capture)
        if (
            hasattr(self.node.configuration, "inputlink_ordering_list")
            and self.node.configuration.inputlink_ordering_list is not None
        ):
            ordering_list = self.node.configuration.inputlink_ordering_list
            inputs_order = []
            for link in ordering_list:
                if "link_label" in link and link["link_label"] == "Before":
                    link_name = link["link_name"] if "link_name" in link else None
                    if link_name:
                        for input in self.inputs:
                            if input.link.get_link_name() == link_name:
                                inputs_order.append(input.id)
            for link in ordering_list:
                if "link_label" in link and link["link_label"] == "After":
                    link_name = link["link_name"] if "link_name" in link else None
                    if link_name:
                        for input in self.inputs:
                            if input.link.get_link_name() == link_name:
                                inputs_order.append(input.id)
            return "|".join(inputs_order)
        else:
            return None

    def get_generic_inputs_order(self):
        assert isinstance(self.node.configuration, generic)
        if (
            hasattr(self.node.configuration, "inputlink_ordering_list")
            and self.node.configuration.inputlink_ordering_list is not None
        ):
            ordering_list = self.node.configuration.inputlink_ordering_list
            inputs_order = []
            for i in range(len(ordering_list)):
                for link in ordering_list:
                    if "link_label" in link and link["link_label"] == str(i):
                        link_name = link["link_name"] if "link_name" in link else None
                        if link_name:
                            for input in self.inputs:
                                if input.link.get_link_name() == link_name:
                                    inputs_order.append(input.id)
            return "|".join(inputs_order)
        else:
            return None

    def get_generic_outputs_order(self):
        assert isinstance(self.node.configuration, generic)
        if (
            hasattr(self.node.configuration, "outputlink_ordering_list")
            and self.node.configuration.outputlink_ordering_list is not None
        ):
            ordering_list = self.node.configuration.outputlink_ordering_list
            outputs_order = []
            for i in range(len(ordering_list)):
                for link in ordering_list:
                    if "link_label" in link and link["link_label"] == str(i):
                        link_name = link["link_name"] if "link_name" in link else None
                        if link_name:
                            for output in self.outputs:
                                if output.link.get_link_name() == link_name:
                                    outputs_order.append(output.id)
            return "|".join(outputs_order)
        else:
            return None

    def get_general_inputs_order(self):
        inputs_order = []
        for input in self.inputs:
            inputs_order.append(input.id)
        return "|".join(inputs_order)

    def get_general_outputs_order(self):
        outputs_order = []
        for output in self.outputs:
            outputs_order.append(output.id)
        return "|".join(outputs_order)

    def validate_cff_level_numbers(self):
        assert isinstance(self.node.configuration, complex_flat_file)
        for record in self.node.configuration.records:
            assert isinstance(record, Record)
            current_level_number = 2
            group = (
                record.columns[0].native_type.value == "GROUP"
                if len(record.columns)
                else False
            )
            for column in record.columns:
                if column.level == current_level_number:
                    continue
                if column.level > current_level_number:
                    if not group:
                        logger.warning(
                            f"\nColumn {column.name} has an invalid level number\n",
                        )
                    current_level_number = column.level
                group = column.native_type.value == "GROUP"
                current_level_number = column.level

    def process_rest_requests(self, requests):
        if not len(requests):
            return requests
        for i, request in enumerate(requests):
            request["id"] = i
            request["label"] = f"Request {i}"
        return requests

    def process_rest_schema(self):
        for output in self.outputs:
            if output.schema:
                for field in output.schema.schema.fields:
                    if "derivation" in field.app_data:
                        field.metadata.description = field.app_data["derivation"]

    def serialize(self):
        """Serializes the execution node into a model.ExecutionNode object."""
        ui_data = self._serialize_ui_data(self.node)
        inputs_order = None
        if self.node.configuration.op_name == "PxJoin":
            inputs_order = self.get_join_inputs_order()
        elif self.node.configuration.op_name == "PxLookup":
            inputs_order = self.get_lookup_inputs_order()
        elif self.node.configuration.op_name == "PxChangeCapture":
            inputs_order = self.get_change_capture_inputs_order()
        elif self.node.configuration.op_name == "PxGeneric":
            inputs_order = self.get_generic_inputs_order()
        elif (
            self.node.configuration.op_name in ["PxFunnel", "XMLStagePX"]
        ):
            inputs_order = self.get_general_inputs_order()

        outputs_order = None
        if self.node.configuration.op_name == "PxLookup":
            outputs_order = self.get_lookup_outputs_order()
        elif self.node.configuration.op_name == "CTransformerStage":
            outputs_order = self.get_transformer_outputs_order()
        elif self.node.configuration.op_name == "PxGeneric":
            outputs_order = self.get_generic_outputs_order()
        elif (
            self.node.configuration.op_name in ["PxJoin", "XMLStagePX"]
        ):
            outputs_order = self.get_general_outputs_order()

        stage_records = None
        if self.node.configuration.op_name == "PxCFF":
            self.validate_cff_level_numbers()
            stage_records = self.get_cff_stage_records()

        stage_records = None
        if self.node.configuration.op_name == "PxCFF":
            self.validate_cff_level_numbers()
            stage_records = self.get_cff_stage_records()

        datastage = {}
        if inputs_order is not None:
            datastage["inputs_order"] = inputs_order
        if outputs_order is not None:
            datastage["outputs_order"] = outputs_order
        if stage_records is not None:
            datastage["stage_records"] = stage_records

        if not len(datastage):
            datastage = None

        parameters = {
            **self.node._get_node_params(),
            **self.node._get_advanced_params(),
        }

        if self.node.configuration.op_name == "CTransformerStage":
            if "Triggers" in parameters:
                for trigger in parameters["Triggers"]:
                    if "arguments" in trigger:
                        args = trigger.pop("arguments")
                        trigger.update(args)

        if self.node.configuration.op_name == "PxColumnImport":
            if "schema" in parameters:
                parameters["schema"] = [
                    col_schema["ColumnName"] for col_schema in parameters["schema"]
                ]

        if self.node.configuration.op_name == "PxRest":
            if "requests" in parameters:
                parameters["requests"] = self.process_rest_requests(
                    parameters["requests"]
                )
            self.process_rest_schema()

        inputs = [ip.serialize() for ip in self.inputs]
        if not len(inputs):
            return model.ExecutionNode(
                id=self.id,
                op=self.node._get_op_name(),
                parameters=parameters,
                outputs=[op.serialize() for op in self.outputs],
                inputs=[
                    {
                        "id": "",
                        "app_data": {
                            "ui_data": {
                                "cardinality": {"min": 0, "max": 1},
                                "label": "inPort",
                            }
                        },
                    }
                ],
                app_data=model.NodeAppData(ui_data=ui_data, datastage=datastage),
            )
        else:
            return model.ExecutionNode(
                id=self.id,
                op=self.node._get_op_name(),
                parameters=parameters,
                inputs=[ip.serialize() for ip in self.inputs],
                outputs=[op.serialize() for op in self.outputs],
                app_data=model.NodeAppData(ui_data=ui_data, datastage=datastage),
            )


class BindingNodeExtractor(AbstractNodeExtractor):
    """Extracts a binding node within a pipeline."""

    def __init__(self, parent_pipe: PipelineExtractor, stage_node: StageNode):
        super().__init__(parent_pipe, stage_node)
        assert isinstance(stage_node, StageNode)
        self.node: StageNode = stage_node

    def get_input_port_params(self) -> dict[str, Any]:
        return self.node._get_input_port_params()

    def get_output_port_params(self) -> dict[str, Any]:
        return self.node._get_output_port_params()

    def get_source_connection_params(self):
        return self.node._get_source_connection_params()

    def get_target_connection_params(self):
        return self.node._get_target_connection_params()

    def serialize(self):
        """Serializes the binding node into a model.BindingEntryNode or model.BindingExitNode object."""
        ui_data = self._serialize_ui_data(self.node)

        # Binding nodes with inputs must be exit nodes (sinks), while those without inputs are entry nodes (sources).
        if self.inputs:
            return model.BindingExitNode(
                id=self.id,
                op=self.node._get_op_name(),
                parameters={
                    **self.node._get_node_params(),
                    **self.node._get_advanced_params(),
                },
                inputs=[ip.serialize() for ip in self.inputs],
                outputs=[op.serialize() for op in self.outputs],
                app_data=model.NodeAppData(ui_data=ui_data),
            )
        else:
            return model.BindingEntryNode(
                id=self.id,
                op=self.node._get_op_name(),
                parameters={
                    **self.node._get_node_params(),
                    **self.node._get_advanced_params(),
                },
                outputs=[op.serialize() for op in self.outputs],
                app_data=model.NodeAppData(ui_data=ui_data),
            )


class ConnectorNodeExtractor(BindingNodeExtractor):
    def __init__(self, parent_pipe: PipelineExtractor, stage_node: StageNode):
        super().__init__(parent_pipe, stage_node)
        self.node: StageNode = stage_node

        if self.node.metadata["out_degree"] > 0 and self.node.metadata["in_degree"] > 0:
            self.location = "both"
        elif self.node.metadata["out_degree"] > 0:
            self.location = "source"
        else:
            self.location = "target"

    def serialize(self):
        mod = super().serialize()
        mod.connection = self.serialize_connection()
        return mod

    def serialize_connection(self):
        """Serializes the connection into a model.CommonPipelineConnection object"""
        # if not hasattr(self.node.configuration, "connection") or (
        #     hasattr(self.node.configuration, "connection") and not self.node.configuration.connection
        # ):
        #     return None
        # return model.CommonPipelineConnection(
        #     name=self.node._get_connection_name(),
        #     ref=self.node._get_connection_id(),
        #     # catalog_ref="",
        #     project_ref=self.node._get_project_id(),
        #     # space_ref="",
        #     properties=self.node._get_connection_params(self.location),
        #     app_data=model.AppDataDef(),
        # )

        if not hasattr(self.node.configuration, "connection") or (
            hasattr(self.node.configuration, "connection")
            and not self.node.configuration.connection
        ):
            return None
        conn = self.node.configuration.connection
        properties = self.node._get_connection_params(self.location)
        ref = self.node._get_connection_id()

        if not ref:
            ref = "local"
            included = [field for field in conn.model_fields_set]
            properties |= json.loads(conn.model_dump_json(include=included, exclude_none=True, by_alias=True, warnings=False))

        return model.CommonPipelineConnection(
            name=self.node._get_connection_name(),
            ref=ref,
            # catalog_ref="",
            project_ref=self.node._get_project_id(),
            # space_ref="",
            properties=properties,
            app_data=model.AppDataDef(),
        )

class ValueDerivationExtractor(AbstractIdent):
    def __init__(
        self, node: AbstractNodeExtractor, schema: SchemaExtractor, link: LinkExtractor
    ):
        super().__init__()
        self.node = node
        self.schema = schema
        self.link = link
        self.configuration = self.node.stage_node.configuration
        self.inputs = self.node.inputs
        self.outputs = self.node.outputs
        self.value_derivation = []

    def get_node_value_derivation(self):
        valueDerivation = []
        op = self.configuration.op_name
        if len(self.inputs) > 1:
            num_primary = 0
            for input in self.inputs:
                if input.link.link.type == "PRIMARY":
                    num_primary += 1
            if num_primary > 1:
                if len(self.outputs) == 1:
                    for output in self.outputs:
                        if output.schema:
                            for field in output.schema.schema.fields:
                                expression = {
                                    "columnName": field.name,
                                    "parsedExpression": field.name,
                                    "sourceColumn": "",
                                }
                                for input in self.inputs:
                                    if input.schema:
                                        for input_field in input.schema.schema.fields:
                                            if input_field.name == field.name:
                                                expression[
                                                    "sourceColumn"
                                                ] += f"{input.link.get_link_name()}.{input_field.name};"
                                if (
                                    expression["sourceColumn"]
                                    and expression["sourceColumn"][-1] == ";"
                                ):
                                    expression["sourceColumn"] = expression[
                                        "sourceColumn"
                                    ][:-1]
                                valueDerivation.append(expression)
        if op == "PxMerge":
            valueDerivation = []
        if op == "PxSort":
            assert isinstance(self.configuration, sort)
            for input in self.inputs:
                link_name = input.link.get_link_name()
                if self.configuration.flag_key.value == "flagKey":
                    column_name = None
                    for output in self.outputs:
                        for field in output.schema.schema.fields:
                            if (
                                "key_change" in field.app_data
                                and field.app_data["key_change"]
                            ):
                                column_name = field.name
                    if not column_name:
                        column_name = "keyChange"

                    valueDerivation.append(
                        {
                            "columnName": column_name,
                            "parsedExpression": "KeyChange()",
                            "sourceColumn": f"{link_name}.keyChange",
                        }
                    )
                if self.configuration.flag_cluster.value == "flagCluster":
                    column_name = None
                    for output in self.outputs:
                        for field in output.schema.schema.fields:
                            if (
                                "cluster_key_change" in field.app_data
                                and field.app_data["cluster_key_change"]
                            ):
                                column_name = field.name
                    if not column_name:
                        column_name = "clusterKeyChange"

                    valueDerivation.append(
                        {
                            "columnName": column_name,
                            "parsedExpression": "ClusterKeyChange()",
                            "sourceColumn": f"{link_name}.{column_name}",
                        }
                    )
        elif op == "PxPeek":
            assert isinstance(self.configuration, peek)
            link_name = self.link.get_link_name()
            if self.configuration.dataset.value == "dataset":
                valueDerivation.append(
                    {
                        "columnName": "rec",
                        "parsedExpression": "Peek()",
                        "sourceColumn": f"{link_name}.rec",
                    }
                )
        elif op == "PxDifference":
            assert isinstance(self.configuration, difference)
            valueDerivation = []
            for input in self.inputs:
                column_name = None
                for output in self.outputs:
                    for field in output.schema.schema.fields:
                        if (
                            "difference" in field.app_data
                            and field.app_data["difference"]
                        ):
                            column_name = field.name
                if not column_name:
                    column_name = "diff"
                valueDerivation.append(
                    {
                        "columnName": column_name,
                        "parsedExpression": "DiffCode()",
                        "sourceColumn": f"{input.link.get_link_name()}.{column_name}",
                    }
                )
        elif op == "PxChecksum":
            assert isinstance(self.configuration, checksum)
            for input in self.inputs:
                link_name = input.link.get_link_name()
                if self.configuration.checksum_name:
                    valueDerivation.append(
                        {
                            "columnName": self.configuration.checksum_name,
                            "parsedExpression": "Checksum()",
                            "sourceColumn": f"{link_name}.{self.configuration.checksum_name}",
                        }
                    )
                else:
                    valueDerivation.append(
                        {
                            "columnName": "checksum",
                            "parsedExpression": "Checksum()",
                            "sourceColumn": f"{link_name}.checksum",
                        }
                    )
                if self.configuration.export_name:
                    valueDerivation.append(
                        {
                            "columnName": self.configuration.export_name,
                            "parsedExpression": "CRCBuffer()",
                            "sourceColumn": f"{link_name}.{self.configuration.export_name}",
                        }
                    )
        elif op == "PxBLM":
            assert isinstance(self.configuration, bloom_filter)
            link_name = self.link.get_link_name()
            valueDerivation.append(
                {
                    "columnName": "dflag",
                    "parsedExpression": "IsADuplicate()",
                    "sourceColumn": f"{link_name}.dflag",
                }
            )
        elif op == "PxChangeCapture":
            assert isinstance(self.configuration, change_capture)
            valueDerivation = []
            link_name = None
            for input_link in self.configuration.inputlink_ordering_list:
                # TODO change when objects (instead of dict)
                if "link_label" in input_link and (
                    input_link["link_label"] == "1"
                    or input_link["link_label"] == "After"
                ):
                    link_name = input_link["link_name"]
            for input in self.inputs:
                new_link_name = link_name or input.link.get_link_name()
                column_name = self.configuration.code_field or "change_code"
                for output in self.outputs:
                    for field in output.schema.schema.fields:
                        if (
                            "change_code" in field.app_data
                            and field.app_data["change_code"]
                        ):
                            column_name = field.name
                valueDerivation.append(
                    {
                        "columnName": column_name,
                        "parsedExpression": "ChangeCode()",
                        "sourceColumn": f"{new_link_name}.change_code",
                    }
                )
                if not link_name:
                    logger.warning(
                        "\nProper input ordering has not been set for Change Capture stage. Input ordering will be randomly assigned.",
                    )
                break
        elif op == "PxChangeApply":
            assert isinstance(self.configuration, change_apply)
            link_name = None
            input_link1 = None
            input_link2 = None
            if len(self.configuration.inputlink_ordering_list) > 1:
                if "link_name" in self.configuration.inputlink_ordering_list[0]:
                    input_link1 = self.configuration.inputlink_ordering_list[0][
                        "link_name"
                    ]
                if "link_name" in self.configuration.inputlink_ordering_list[1]:
                    input_link2 = self.configuration.inputlink_ordering_list[1][
                        "link_name"
                    ]
                for input_link in self.configuration.inputlink_ordering_list:
                    if (
                        "link_label" in input_link
                        and input_link["link_label"] == "Change"
                    ):
                        if "link_name" in input_link:
                            link_name = input_link["link_name"]
            change_schema = None
            for input_node in self.inputs:
                if input_node.link.get_link_name() == link_name:
                    change_schema = input_node.schema.schema
            change_code_column_name = self.configuration.code_field or "change_code"
            if change_schema and len(change_schema.fields) != 0:
                for field in change_schema.fields:
                    column_name = field.name
                    if column_name and column_name != change_code_column_name:
                        valueDerivation.append(
                            {
                                "columnName": column_name,
                                "parsedExpression": f"{link_name}.{column_name}",
                                "sourceColumn": f"{input_link1}.{column_name};{input_link2}.{column_name}",
                            }
                        )
        elif op == "PxCompare":
            assert isinstance(self.configuration, compare)
            valueDerivation = []
            # valueDerivation.append(
            #     {"columnName": "first", "parsedExpression": "-", "sourceColumn": "-"}
            # )
            # valueDerivation.append(
            #     {"columnName": "second", "parsedExpression": "-", "sourceColumn": "-"}
            # )
            # valueDerivation.append(
            #     {"columnName": "result", "parsedExpression": "-", "sourceColumn": "-"}
            # )
        elif op == "MatchFrequency":
            assert isinstance(self.configuration, match_frequency)
            mf_columns = [
                "qsFreqValue",
                "qsFreqCounts",
                "qsFreqColumnID",
                "qsFreqHeaderFlag",
            ]
            for input in self.inputs:
                link_name = input.link.get_link_name()
                for derivation in mf_columns:
                    valueDerivation.append(
                        {
                            "columnName": derivation,
                            "parsedExpression": f"MatchFrequency({derivation})",
                            "sourceColumn": f"{link_name}.{derivation}",
                        }
                    )
        elif op == "UnduplicateMatch":
            mf_columns = [
                "qsMatchType",
                "qsMatchWeight",
                "qsMatchPattern",
                "qsMatchLRFlag",
                "qsMatchExactFlag",
                "qsMatchPassNumber",
                "qsMatchSetID",
                "qsMatchDataID",
                "qsMatchStatType",
                "qsMatchStatValue",
            ]
            for input in self.inputs:
                link_name = input.link.get_link_name()
                for derivation in mf_columns:
                    valueDerivation.append(
                        {
                            "columnName": derivation,
                            "parsedExpression": f"UnduplicateMatch({derivation})",
                            "sourceColumn": f"{link_name}.{derivation}",
                        }
                    )
        elif op == "ReferenceMatch":
            mf_columns = [
                "qsMatchRefID",
                "qsMatchType",
                "qsMatchWeight",
                "qsMatchPattern",
                "qsMatchLRFlag",
                "qsMatchExactFlag",
                "qsMatchPassNumber",
                "qsMatchSetID",
                "qsMatchDataID",
                "qsMatchStatType",
                "qsMatchStatValue",
            ]
            for input in self.inputs:
                link_name = input.link.get_link_name()
                for derivation in mf_columns:
                    valueDerivation.append(
                        {
                            "columnName": derivation,
                            "parsedExpression": f"ReferenceMatch({derivation})",
                            "sourceColumn": f"{link_name}.{derivation}",
                        }
                    )
        elif op in [
            "QSmwi",
        ]:
            pass
        elif op == "AddressVerification2":
            assert isinstance(self.configuration, address_verification)
            column_data = {
                "AccuracyCode_QSAV": "AccuracyCode",
                "AddressQualityIndex_QSAV": "AddressQualityIndex",
                "Organization_QSAV": "Organization",
                "Department_QSAV": "Department",
                "Function_QSAV": "Function",
                "Contact_QSAV": "Contact",
                "Building_QSAV": "Building",
                "Subbuilding_QSAV": "Subbuilding",
                "HouseNumber_QSAV": "HouseNumber",
                "Street_QSAV": "Street",
                "DependentStreet_QSAV": "DependentStreet",
                "POBOX_QSAV": "POBOX",
                "Locality_QSAV": "Locality",
                "DependentLocality_QSAV": "DependentLocality",
                "DoubleDependentLocality_QSAV": "DoubleDependentLocality",
                "PostCode_QSAV": "PostCode",
                "PostalCodePrimary_QSAV": "PostalCodePrimary",
                "PostalCodeSecondary_QSAV": "PostalCodeSecondary",
                "SuperAdministrativeArea_QSAV": "SuperAdministrativeArea",
                "AdministrativeArea_QSAV": "AdministrativeArea",
                "SubAdministrativeArea_QSAV": "SubAdministrativeArea",
                "Country_QSAV": "Country",
                "ISO3166_2_QSAV": "ISO3166",
                "ISO3166_3_QSAV": "ISO3166",
                "ISO3166_N_QSAV": "ISO3166",
                "Address_QSAV": "Address",
                "Residue_QSAV": "Residue",
                "DeliveryAddress1_QSAV": "DeliveryAddress1",
                "DeliveryAddress2_QSAV": "DeliveryAddress2",
                "DeliveryAddress3_QSAV": "DeliveryAddress3",
                "DeliveryAddress4_QSAV": "DeliveryAddress4",
                "DeliveryAddress5_QSAV": "DeliveryAddress5",
                "DeliveryAddress6_QSAV": "DeliveryAddress6",
                "DeliveryAddress7_QSAV": "DeliveryAddress7",
                "DeliveryAddress8_QSAV": "DeliveryAddress8",
                "FormattedAddress1_QSAV": "FormattedAddress1",
                "FormattedAddress2_QSAV": "FormattedAddress2",
                "FormattedAddress3_QSAV": "FormattedAddress3",
                "FormattedAddress4_QSAV": "FormattedAddress4",
                "FormattedAddress5_QSAV": "FormattedAddress5",
                "FormattedAddress6_QSAV": "FormattedAddress6",
                "FormattedAddress7_QSAV": "FormattedAddress7",
                "FormattedAddress8_QSAV": "FormattedAddress8",
                "FormattedAddress9_QSAV": "FormattedAddress9",
                "FormattedAddress10_QSAV": "FormattedAddress10",
                "AddressFormat_QSAV": "AddressFormat",
                "DeliveryAddress_QSAV": "DeliveryAddress",
                "DeliveryAddressFormat_QSAV": "DeliveryAddressFormat",
                "LocalityExtra_QSAV": "LocalityExtra",
                "LocalitySpecial_QSAV": "LocalitySpecial",
                "PremiseExtra_QSAV": "PremiseExtra",
                "ErrorCode_QSAV": "ErrorCode",
                "ErrorMessage_QSAV": "ErrorMessage",
            }
            for field in self.schema.schema.fields:
                if field.name in column_data:
                    valueDerivation.append(
                        {
                            "columnName": field.name,
                            "parsedExpression": column_data[field.name],
                            "sourceColumn": f".{field.name}",
                        }
                    )

        elif op == "PxFunnel":
            valueDerivation = []
            assert isinstance(self.configuration, funnel)
            output_col = []
            result = []
            for output_node in self.outputs:
                if output_node.schema:
                    output_col.append(output_node.schema.schema.fields)
            if len(output_col) > 0:
                result = output_col.pop(0)
                result = [
                    v
                    for v in result
                    if all(any(x.name == v.name for x in a) for a in output_col)
                ]
            # if result:
            #     for col in result:
            #         source_column = ""
            #         for input_node in self.inputs:
            #             if input_node.link.get_link_name():
            #                 append_str = ";"
            #                 if source_column != "":
            #                     source_column += (
            #                         append_str
            #                         + f"{input_node.link.get_link_name()}.{col.name}"
            #                     )
            #                 else:
            #                     source_column = (
            #                         f"{input_node.link.get_link_name()}.{col.name}"
            #                     )
            #         valueDerivation.append(
            #             {
            #                 "columnName": col.name,
            #                 "parsedExpression": col.name,
            #                 "sourceColumn": source_column,
            #             }
            #         )
            for output_node in self.outputs:
                if not output_node.schema:
                    continue
                for field in output_node.schema.schema.fields:
                    # result_names = [res.name for res in result]
                    # if field.name not in result_names:
                    source_column = ""
                    parsed_expression = field.metadata.source_field_id or field.name
                    field.metadata.source_field_id = None
                    for input_node in self.inputs:
                        if input_node.link.get_link_name():
                            append_str = ";"
                            if source_column != "":
                                source_column += (
                                    append_str
                                    + f"{input_node.link.get_link_name()}.{parsed_expression}"
                                )
                            else:
                                source_column = f"{input_node.link.get_link_name()}.{parsed_expression}"

                    valueDerivation.append(
                        {
                            "columnName": field.name,
                            "parsedExpression": parsed_expression,
                            "sourceColumn": source_column,
                        }
                    )
        elif op == "PxAggregator":
            assert isinstance(self.configuration, aggregator)
            selection = self.configuration.selection.value
            operations = {
                "sum": "Sum",
                "min": "Min",
                "max": "Max",
                "css": "CorrectedSumSquares",
                "mean": "Mean",
                "missing": "MissingValCount",
                "count": "Count",
                "cv": "PctCoeffVar",
                "range": "Range",
                "std": "StandardDeviation",
                "ste": "StandardError",
                "sumw": "SumWeights",
                "uss": "UncorrectedSumSquares",
                "var": "Variance",
            }
            if len(self.configuration.reduce_properties) > 0:
                for input in self.inputs:
                    for reduce_props in self.configuration.reduce_properties:
                        selected_column = None
                        for reduce_prop in reduce_props:
                            if selection in reduce_prop:
                                selected_column = reduce_prop[selection]
                        for reduce_prop in reduce_props:
                            summary_operations = {
                                "n": "SummaryNumRecs",
                                "nMissing": "SummaryNumMissingVals",
                                "sumOfWeights": "SummarySumWeights",
                                "minimum": "SummaryMinimum",
                                "maximum": "SummaryMaximum",
                                "mean": "SummaryMean",
                                "css": "SummarySumSquares",
                            }
                            for agg_op in operations:
                                if agg_op in reduce_prop:
                                    column_name = None
                                    for output in self.outputs:
                                        for field in output.schema.schema.fields:
                                            if field.name == reduce_prop[agg_op]:
                                                column_name = reduce_prop[agg_op]
                                    if not column_name:
                                        column_name = selected_column
                                    valueDerivation.append(
                                        {
                                            "columnName": column_name,
                                            "parsedExpression": f"{operations[agg_op]}({input.link.get_link_name()}.{selected_column})",
                                            "sourceColumn": f"{input.link.get_link_name()}.{selected_column}",
                                        }
                                    )
                            if "summary" in reduce_prop:
                                column_name = None
                                for output in self.outputs:
                                    for field in output.schema.schema.fields:
                                        if field.name == reduce_prop["summary"]:
                                            column_name = reduce_prop["summary"]
                                if not column_name:
                                    column_name = selected_column
                                valueDerivation.append(
                                    {
                                        "columnName": column_name,
                                        "parsedExpression": f"Summary({input.link.get_link_name()}.{selected_column})",
                                        "sourceColumn": f"{input.link.get_link_name()}.{selected_column}",
                                    }
                                )
                                for suffix, operation in summary_operations.items():
                                    valueDerivation.append(
                                        {
                                            "columnName": f"{column_name}.{suffix}",
                                            "parsedExpression": f"{operation}({input.link.get_link_name()}.{selected_column})",
                                            "sourceColumn": f"{input.link.get_link_name()}.{selected_column}",
                                        }
                                    )

            if len(self.configuration.rereduce_properties) > 0:
                for input in self.inputs:
                    for rereduce_props in self.configuration.rereduce_properties:
                        selected_column = None
                        for rereduce_prop in rereduce_props:
                            if selection in rereduce_prop:
                                selected_column = rereduce_prop[selection]
                        for rereduce_prop in rereduce_props:
                            for agg_op in operations:
                                if agg_op in rereduce_prop:
                                    column_name = None
                                    for output in self.outputs:
                                        for field in output.schema.schema.fields:
                                            if field.name == rereduce_prop[agg_op]:
                                                column_name = rereduce_prop[agg_op]
                                    if not column_name:
                                        column_name = selected_column
                                    valueDerivation.append(
                                        {
                                            "columnName": column_name,
                                            "parsedExpression": f"{operations[agg_op]}({input.link.get_link_name()}.{selected_column})",
                                            "sourceColumn": f"{input.link.get_link_name()}.{selected_column}",
                                        }
                                    )

            if len(self.configuration.count_field_properties) > 0:
                for input in self.inputs:
                    for count_fields in self.configuration.count_field_properties:
                        for count_field in count_fields:
                            if "countField" in count_field:
                                valueDerivation.append(
                                    {
                                        "columnName": count_field["countField"],
                                        "parsedExpression": "RecCount()",
                                        "sourceColumn": f"{input.link.get_link_name()}.{count_field['countField']}",
                                    }
                                )
        elif op == "PxColumnImport":
            assert isinstance(self.configuration, column_import)
            input_link = None
            if self.configuration.field:
                for input_node in self.inputs:
                    for field in input_node.schema.schema.fields:
                        if field.name == self.configuration.field:
                            input_link = input_node.link.link.name
                            break

                if self.link.link.type == "PRIMARY":
                    for field in self.schema.schema.fields:
                        for col_schema in self.configuration.schema_:
                            if (
                                "ColumnName" in col_schema
                                and col_schema["ColumnName"] == field.name
                            ):
                                valueDerivation.append(
                                    {
                                        "columnName": field.name,
                                        "parsedExpression": f"ColumnImport({input_link}.{self.configuration.field},{field.name})",
                                        "sourceColumn": f"{input_link}.{self.configuration.field}",
                                    }
                                )
        elif op == "PxColumnExport":
            assert isinstance(self.configuration, column_export)
            input_columns = []
            export_columns = self.configuration.schema_
            if self.configuration.field and export_columns:
                for input_node in self.inputs:
                    for field in input_node.schema.schema.fields:
                        if field.name in export_columns:
                            input_columns.append(
                                f"{input_node.link.get_link_name()}.{field.name}"
                            )
            valueDerivation.append(
                {
                    "columnName": self.configuration.field,
                    "parsedExpression": f"ColumnExport({','.join(input_columns)},{self.configuration.field})",
                    "sourceColumn": ";".join(input_columns),
                }
            )
        elif op == "PxColumnGenerator":
            assert isinstance(self.configuration, column_generator)
            input_link = None
            if len(self.configuration.schema_) > 0:
                for schema in self.configuration.schema_:
                    for input in self.inputs:
                        valueDerivation.append(
                            {
                                "columnName": schema,
                                "parsedExpression": f"Generate({schema})",
                                "sourceColumn": f"{input.link.get_link_name()}.{schema}",
                            }
                        )
        elif op == "PxPivot":
            assert isinstance(self.configuration, pivot)
            if (
                hasattr(self.configuration.pivot_type, "value")
                and self.configuration.pivot_type.value == "verticalpivot"
            ) or self.configuration.pivot_type == "verticalpivot":
                for input in self.inputs:
                    for output in self.outputs:
                        for field in output.schema.schema.fields:
                            valueDerivation.append(
                                {
                                    "columnName": field.name,
                                    "parsedExpression": f"VerticalPivot({field.name})",
                                    "sourceColumn": f"{input.link.get_link_name()}.{field.name}",
                                }
                            )
            elif self.configuration.pivot_properties:
                for input in self.inputs:
                    for output in self.outputs:
                        for field in output.schema.schema.fields:
                            for pivot_prop in self.configuration.pivot_properties:
                                if "name" in pivot_prop:
                                    if pivot_prop["name"] == field.name or (
                                        "pivot_property" in field.app_data
                                        and field.app_data["pivot_property"]
                                        == pivot_prop["name"]
                                    ):
                                        valueDerivation.append(
                                            {
                                                "columnName": field.name,
                                                "parsedExpression": f"Pivot({pivot_prop['name']})",
                                                "sourceColumn": f"{input.link.get_link_name()}.{field.name}",
                                            }
                                        )
        elif op == "Standardize":
            assert isinstance(self.configuration, standardize)
            ruleset = self.configuration.ruleset_properties
            for output in self.outputs:
                for field in output.schema.schema.fields:
                    for rule in ruleset:
                        if "ruleset" in rule and field.name.endswith(rule["ruleset"]):
                            for input in self.inputs:
                                valueDerivation.append(
                                    {
                                        "columnName": field.name,
                                        "parsedExpression": f"Standardize({field.name})",
                                        "sourceColumn": f"{input.link.get_link_name()}.{field.name}",
                                    }
                                )
        elif op == "Investigate":
            investigate_names = [
                "qsInvColumnName",
                "qsInvPattern",
                "qsInvSample",
                "qsInvCount",
                "qsInvPercent",
                "qsInvWord",
                "qsInvClassCode",
            ]
            for output in self.outputs:
                for field in output.schema.schema.fields:
                    if field.name in investigate_names:
                        for input in self.inputs:
                            valueDerivation.append(
                                {
                                    "columnName": field.name,
                                    "parsedExpression": f"Investigate({field.name})",
                                    "sourceColumn": f"{input.link.get_link_name()}.{field.name}",
                                }
                            )
        elif op == "XMLStagePX" or op == "PxJoin":
            valueDerivation = []
        elif op == "CTransformerStage":
            assert isinstance(self.configuration, transformer)
            if self.link.link.schema.configuration:
                for field in self.link.link.schema.configuration.fields:
                    if "derivation" in field.app_data:
                        source_column = field.metadata.source_field_id
                        if not source_column:
                            found_field = None
                            num_found = 0
                            for input in self.inputs:
                                if input.schema:
                                    for input_field in input.schema.schema.fields:
                                        if input_field.name == field.name:
                                            found_field = f"{input.link.get_link_name()}.{input_field.name}"
                                            num_found += 1
                            if num_found == 1:
                                source_column = found_field
                        elif "." not in source_column:
                            found_field = None
                            num_found = 0
                            for input in self.inputs:
                                if input.schema:
                                    for input_field in input.schema.schema.fields:
                                        if input_field.name == source_column:
                                            found_field = f"{input.link.get_link_name()}.{input_field.name}"
                                            num_found += 1
                            if num_found == 1:
                                source_column = found_field

                        if source_column:
                            valueDerivation.append(
                                {
                                    "columnName": field.name,
                                    "parsedExpression": field.app_data["derivation"],
                                    "sourceColumn": source_column,
                                }
                            )
                        else:
                            valueDerivation.append(
                                {
                                    "columnName": field.name,
                                    "parsedExpression": field.app_data["derivation"],
                                }
                            )

        self.value_derivation.extend(valueDerivation)

        if self.schema:
            for field in self.schema.schema.fields:
                found = False
                for vd in self.value_derivation:
                    if "columnName" in vd and vd["columnName"] == field.name:
                        found = True
                if not found:
                    if not field.metadata.source_field_id:
                        num_found = 0
                        link_name = None
                        for input_node in self.inputs:
                            if input_node.schema:
                                for input_field in input_node.schema.schema.fields:
                                    if field.name == input_field.name:
                                        num_found += 1
                                        link_name = input_node.link.get_link_name()
                        if num_found == 1:
                            field.metadata.source_field_id = f"{link_name}.{field.name}"
                    elif "." not in field.metadata.source_field_id:
                        num_found = 0
                        link_name = None
                        for input_node in self.inputs:
                            if input_node.schema:
                                for input_field in input_node.schema.schema.fields:
                                    if (
                                        field.metadata.source_field_id
                                        == input_field.name
                                    ):
                                        num_found += 1
                                        link_name = input_node.link.get_link_name()
                        if num_found == 1:
                            field.metadata.source_field_id = (
                                f"{link_name}.{field.metadata.source_field_id}"
                            )

    def get_link_value_derivation(self):
        pass

    def serialize(self):
        self.get_node_value_derivation()
        return self.value_derivation


class OutputPortExtractor(AbstractIdent):
    """Extracts an output port for a node."""

    def __init__(
        self,
        parent_pipe: PipelineExtractor,
        link: LinkExtractor,
        node: AbstractNodeExtractor,
        schema: SchemaExtractor | None,
    ):
        super().__init__()
        self.parent_pipe = parent_pipe
        self.link = link
        self.node_ext = node
        self.schema = schema

    def serialize(self):
        """Serializes the output port into a model.Port object."""
        cardinality = (
            self.node_ext.stage_node.configuration.get_output_cardinality()
        )

        app_data = model.PortAppData(
            ui_data=model.PortUI(
                label="outPort",
                cardinality=cardinality,
            ),
        )
        app_data.datastage = {"is_source_of_link": self.link.id}

        value_derivation = None
        if isinstance(self.node_ext.stage_node, StageNode):
            app_data.ui_data = model.PortUI(
                label="outPort",
                cardinality=self.node_ext.stage_node.configuration.get_output_cardinality(),
            )
            vd_extractor = ValueDerivationExtractor(
                self.node_ext, self.schema, self.link
            )
            value_derivation = vd_extractor.serialize()

        if (
            self.node_ext.stage_node._get_node_label() != None
            and "Java_Integration" in self.node_ext.stage_node._get_node_label()
        ):
            parameters = self.node_ext.get_output_port_params(self.link.get_link_name())
        else:
            parameters = self.node_ext.get_output_port_params()

        if value_derivation:
            parameters["valueDerivation"] = value_derivation

        if (
            hasattr(self.node_ext.stage_node, "configuration")
            and self.node_ext.stage_node.configuration.op_name == "PxSequentialFile"
            and self.link.link.type == "REJECT"
        ):
            parameters["is_reject_output"] = True

        if (
            hasattr(self.node_ext.stage_node, "configuration")
            and self.node_ext.stage_node.configuration.op_name == "CTransformerStage"
        ):
            if "TransformerConstraint" in parameters:
                update = {}
                for constraint in parameters["TransformerConstraint"]:
                    if (
                        "output_name" in constraint
                        and constraint["output_name"] == self.link.get_link_name()
                    ):
                        update["TransformerConstraint"] = (
                            constraint["TransformerConstraint"]
                            if "TransformerConstraint" in constraint
                            else None
                        )
                        update["Reject"] = (
                            constraint["Reject"] if "Reject" in constraint else None
                        )
                        update["RowLimit"] = (
                            constraint["RowLimit"] if "RowLimit" in constraint else None
                        )
                del parameters["TransformerConstraint"]
                parameters.update(update)
        if (
            hasattr(self.node_ext.stage_node, "configuration")
            and self.node_ext.stage_node.configuration.op_name == "PxCFF"
        ):
            if "predicate" in parameters:
                predicate = None
                for constraint in parameters["predicate"]:
                    if (
                        "output_name" in constraint
                        and constraint["output_name"] == self.link.get_link_name()
                    ):
                        predicate = (
                            constraint["constraint"]
                            if "constraint" in constraint
                            else None
                        )
                if predicate:
                    parameters["predicate"] = predicate
                else:
                    del parameters["predicate"]

        if (
            self.node_ext.stage_node.metadata["in_degree"] > 0
            and self.node_ext.stage_node.metadata["out_degree"] > 0
        ):
            app_data.additionalProperties = {
                "enableAcp": self.node_ext.stage_node._get_acp()
            }

            return model.Port(
                id=self.id,
                parameters=parameters,
                schema_ref=self.schema.id if self.schema else None,
                app_data=app_data,
            )
        else:
            return model.Port(
                id=self.id,
                parameters=parameters,
                schema_ref=self.schema.id if self.schema else None,
                app_data=app_data,
            )


class PlainOutputPortExtractor:
    def __init__(self, node: AbstractNodeExtractor):
        self.node = node

    def serialize(self):
        return model.Port(
            id="",
            app_data=model.PortAppData(
                ui_data=model.PortUI(
                    label="outPort",
                    cardinality=model.Cardinality(
                        min=self.node.stage_node._get_min_primary_outputs(),
                        max=self.node.stage_node._get_max_primary_outputs(),
                    ),
                )
            ),
        )

class InputPortExtractor(AbstractIdent):
    """Extracts an input port for a node."""

    def __init__(
        self,
        parent_pipe: PipelineExtractor,
        link: LinkExtractor,
        node: AbstractNodeExtractor,
        schema: SchemaExtractor = None,
    ):
        super().__init__()
        self.parent_pipe = parent_pipe
        self.link = link
        self.node = node
        self.schema = schema

    def serialize(self):
        """Serializes the input port into a model.Port object."""
        if (
            self.node.stage_node._get_node_label() != None
            and "Java_Integration" in self.node.stage_node._get_node_label()
        ):
            parameters = self.node.get_input_port_params(self.link.get_link_name())
        else:
            parameters = self.node.get_input_port_params()
        if (
            hasattr(self.node.node, "configuration")
            and self.node.node.configuration.op_name == "PxLookup"
        ):
            if "lookupDerivation" in parameters:
                lookup_derivation = []
                primary_link = None
                for link in self.node.inputs:
                    if link.link.link.type == "PRIMARY":
                        primary_link = link.link.get_link_name()
                for ld_props in parameters["lookupDerivation"]:
                    if self.link.get_link_name() == ld_props["reference_link"]:
                        for derivation in ld_props["derivations"]:
                            key_expression = (
                                derivation["key_expression"]
                                if "key_expression" in derivation
                                else ""
                            )
                            parsed_expression = None
                            if "key_expression" in derivation:
                                if "Range" in derivation["key_expression"]:
                                    parsed_expression = derivation["key_expression"]
                                else:
                                    parsed_expression = (
                                        f"{primary_link}.{key_expression}"
                                    )
                            lookup_derivation.append(
                                {
                                    "parsedExpression": parsed_expression,
                                    "sourceColumn": f"{primary_link}.{key_expression}",
                                    "keyType": derivation["key_type"],
                                    "columnName": derivation["key_column"],
                                }
                            )
                        if "conditionNotMet" in ld_props:
                            parameters["conditionNotMet"] = ld_props["conditionNotMet"]
                        if "Condition" in ld_props:
                            parameters["Condition"] = ld_props["Condition"]
                        if "lookupFail" in ld_props:
                            parameters["lookupFail"] = ld_props["lookupFail"]

                if lookup_derivation:
                    parameters["lookupDerivation"] = lookup_derivation
                else:
                    del parameters["lookupDerivation"]

        if hasattr(self.link.link.src, "configuration") and hasattr(
            self.link.link.src.configuration, "runtime_column_propagation"
        ):
            if (
                hasattr(self.node.node, "configuration")
                and self.node.node.configuration.op_name == "WSTransformerPX"
            ):
                parameters["runtime_column_propagation"] = int(
                    self.link.link.src.configuration.runtime_column_propagation
                )
            else:
                parameters["runtime_column_propagation"] = bool(
                    self.link.link.src.configuration.runtime_column_propagation
                )
        elif hasattr(self.link.link.src, "rcp"):
            parameters["runtime_column_propagation"] = self.link.link.src.rcp

        return model.Port(
            id=self.id,
            schema_ref=self.schema.id if self.schema else None,
            parameters=parameters,
            links=[self.link.serialize()],
            app_data=model.PortAppData(ui_data=model.PortUI(label="")),
        )



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import SAPBULKEXTRACT_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class ConnectionType(Enum):
    application_server = "application_server"
    load_balancing = "load_balancing"
    snc_application_server = "snc_application_server"
    snc_load_balancing = "snc_load_balancing"


class SncQop(Enum):
    auth_only = "1"
    auth_integrity = "2"
    auth_integrity_privacy = "3"
    global_default = "8"
    max_protection = "9"


class SapbulkextractConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "ec42cd4a-18ab-469c-bfd7-f6dc52e1db22"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    client_number: str = Field(None, alias="client_number")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    connection_type: SAPBULKEXTRACT_CONNECTION.ConnectionType = Field(
        SAPBULKEXTRACT_CONNECTION.ConnectionType.application_server, alias="connection_type"
    )
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    gateway_url: str | None = Field(None, alias="gateway_url")
    group: str = Field(None, alias="group")
    enable_x_509_certificate: bool | None = Field(False, alias="isx509enabled")
    jar_uris: str | None = Field(None, alias="jar_uris")
    language: str = Field(None, alias="language")
    message_server: str = Field(None, alias="message_server")
    sap_password: str | None = Field(None, alias="password")
    application_server: str = Field(None, alias="sap_application_server")
    sap_application_system_number: str = Field(None, alias="sap_application_system_number")
    sap_router: str | None = Field(None, alias="sap_router")
    snc_name: str = Field(None, alias="snc_name")
    snc_partner_name: str = Field(None, alias="snc_partner_name")
    snc_qop: SAPBULKEXTRACT_CONNECTION.SncQop | None = Field(
        SAPBULKEXTRACT_CONNECTION.SncQop.auth_integrity_privacy, alias="snc_qop"
    )
    system_id: str = Field(None, alias="system_id")
    system_number: str = Field(None, alias="system_number")
    use_system_number: bool | None = Field(True, alias="use_system_number")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    x_509_certificate: str = Field(None, alias="x509_cert")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        include.add("client_number") if (not self.defer_credentials) else exclude.add("client_number")
        (
            include.add("system_number")
            if (
                ((self.use_system_number))
                and (
                    (
                        (
                            self.connection_type
                            and (
                                (
                                    hasattr(self.connection_type, "value")
                                    and self.connection_type.value == "load_balancing"
                                )
                                or (self.connection_type == "load_balancing")
                            )
                        )
                    )
                    or (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_load_balancing"
                            )
                            or (self.connection_type == "snc_load_balancing")
                        )
                    )
                )
            )
            else exclude.add("system_number")
        )
        (
            include.add("use_system_number")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (hasattr(self.connection_type, "value") and self.connection_type.value == "load_balancing")
                            or (self.connection_type == "load_balancing")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("use_system_number")
        )
        (
            include.add("system_id")
            if (
                (((not self.use_system_number)))
                and (
                    (
                        (
                            self.connection_type
                            and (
                                (
                                    hasattr(self.connection_type, "value")
                                    and self.connection_type.value == "load_balancing"
                                )
                                or (self.connection_type == "load_balancing")
                            )
                        )
                    )
                    or (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_load_balancing"
                            )
                            or (self.connection_type == "snc_load_balancing")
                        )
                    )
                )
            )
            else exclude.add("system_id")
        )
        (
            include.add("sap_application_system_number")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "application_server"
                            )
                            or (self.connection_type == "application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value == "snc_application_server"
                        )
                        or (self.connection_type == "snc_application_server")
                    )
                )
            )
            else exclude.add("sap_application_system_number")
        )
        include.add("language") if (not self.defer_credentials) else exclude.add("language")
        (
            include.add("enable_x_509_certificate")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_application_server"
                            )
                            or (self.connection_type == "snc_application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("enable_x_509_certificate")
        )
        (
            include.add("snc_partner_name")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_application_server"
                            )
                            or (self.connection_type == "snc_application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("snc_partner_name")
        )
        (
            include.add("snc_name")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_application_server"
                            )
                            or (self.connection_type == "snc_application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("snc_name")
        )
        (
            include.add("sap_password")
            if (
                ((not self.defer_credentials))
                and (
                    (
                        (
                            (
                                self.connection_type
                                and (
                                    (
                                        hasattr(self.connection_type, "value")
                                        and self.connection_type.value == "application_server"
                                    )
                                    or (self.connection_type == "application_server")
                                )
                            )
                        )
                        or (
                            (
                                self.connection_type
                                and (
                                    (
                                        hasattr(self.connection_type, "value")
                                        and self.connection_type.value == "load_balancing"
                                    )
                                    or (self.connection_type == "load_balancing")
                                )
                            )
                        )
                    )
                    or (not self.enable_x_509_certificate)
                )
            )
            else exclude.add("sap_password")
        )
        (
            include.add("application_server")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "application_server"
                            )
                            or (self.connection_type == "application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value == "snc_application_server"
                        )
                        or (self.connection_type == "snc_application_server")
                    )
                )
            )
            else exclude.add("application_server")
        )
        (
            include.add("message_server")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (hasattr(self.connection_type, "value") and self.connection_type.value == "load_balancing")
                            or (self.connection_type == "load_balancing")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("message_server")
        )
        (
            include.add("x_509_certificate")
            if (
                ((self.enable_x_509_certificate))
                and (
                    (
                        (
                            self.connection_type
                            and (
                                (
                                    hasattr(self.connection_type, "value")
                                    and self.connection_type.value == "snc_application_server"
                                )
                                or (self.connection_type == "snc_application_server")
                            )
                        )
                    )
                    or (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_load_balancing"
                            )
                            or (self.connection_type == "snc_load_balancing")
                        )
                    )
                )
            )
            else exclude.add("x_509_certificate")
        )
        (
            include.add("snc_qop")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "snc_application_server"
                            )
                            or (self.connection_type == "snc_application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("snc_qop")
        )
        (
            include.add("sap_router")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (
                                hasattr(self.connection_type, "value")
                                and self.connection_type.value == "application_server"
                            )
                            or (self.connection_type == "application_server")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "load_balancing")
                        or (self.connection_type == "load_balancing")
                    )
                )
                or (
                    self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value == "snc_application_server"
                        )
                        or (self.connection_type == "snc_application_server")
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("sap_router")
        )
        (
            include.add("group")
            if (
                (
                    (
                        self.connection_type
                        and (
                            (hasattr(self.connection_type, "value") and self.connection_type.value == "load_balancing")
                            or (self.connection_type == "load_balancing")
                        )
                    )
                )
                or (
                    self.connection_type
                    and (
                        (hasattr(self.connection_type, "value") and self.connection_type.value == "snc_load_balancing")
                        or (self.connection_type == "snc_load_balancing")
                    )
                )
            )
            else exclude.add("group")
        )
        (
            include.add("username")
            if (
                ((not self.defer_credentials))
                and (
                    (
                        (
                            (
                                self.connection_type
                                and (
                                    (
                                        hasattr(self.connection_type, "value")
                                        and self.connection_type.value == "application_server"
                                    )
                                    or (self.connection_type == "application_server")
                                )
                            )
                        )
                        or (
                            (
                                self.connection_type
                                and (
                                    (
                                        hasattr(self.connection_type, "value")
                                        and self.connection_type.value == "load_balancing"
                                    )
                                    or (self.connection_type == "load_balancing")
                                )
                            )
                        )
                    )
                    or (not self.enable_x_509_certificate)
                )
            )
            else exclude.add("username")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )

        (
            include.add("application_server")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "application_server" in str(self.connection_type.value)
                    )
                    or ("application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
            )
            else exclude.add("application_server")
        )
        (
            include.add("system_id")
            if (self.use_system_number == "false" or not self.use_system_number)
            and (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                and self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("system_id")
        )
        (
            include.add("use_system_number")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("use_system_number")
        )
        include.add("client_number") if (not self.defer_credentials) else exclude.add("client_number")
        (
            include.add("system_number")
            if (self.use_system_number == "true" or self.use_system_number)
            and (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                and self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("system_number")
        )
        (
            include.add("enable_x_509_certificate")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("enable_x_509_certificate")
        )
        (
            include.add("snc_partner_name")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("snc_partner_name")
        )
        (
            include.add("sap_application_system_number")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "application_server" in str(self.connection_type.value)
                    )
                    or ("application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
            )
            else exclude.add("sap_application_system_number")
        )
        (
            include.add("sap_password")
            if (not self.defer_credentials)
            and (
                (
                    self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value
                            and "application_server" in str(self.connection_type.value)
                        )
                        or ("application_server" in str(self.connection_type))
                    )
                    or self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value
                            and "load_balancing" in str(self.connection_type.value)
                        )
                        or ("load_balancing" in str(self.connection_type))
                    )
                )
                or (self.enable_x_509_certificate == "false" or not self.enable_x_509_certificate)
            )
            else exclude.add("sap_password")
        )
        (
            include.add("username")
            if (not self.defer_credentials)
            and (
                (
                    self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value
                            and "application_server" in str(self.connection_type.value)
                        )
                        or ("application_server" in str(self.connection_type))
                    )
                    or self.connection_type
                    and (
                        (
                            hasattr(self.connection_type, "value")
                            and self.connection_type.value
                            and "load_balancing" in str(self.connection_type.value)
                        )
                        or ("load_balancing" in str(self.connection_type))
                    )
                )
                or (self.enable_x_509_certificate == "false" or not self.enable_x_509_certificate)
            )
            else exclude.add("username")
        )
        (
            include.add("snc_qop")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("snc_qop")
        )
        (
            include.add("x_509_certificate")
            if (self.enable_x_509_certificate == "true" or self.enable_x_509_certificate)
            and (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                and self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("x_509_certificate")
        )
        include.add("language") if (not self.defer_credentials) else exclude.add("language")
        (
            include.add("message_server")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("message_server")
        )
        (
            include.add("sap_router")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "application_server" in str(self.connection_type.value)
                    )
                    or ("application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("sap_router")
        )
        (
            include.add("group")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "load_balancing" in str(self.connection_type.value)
                    )
                    or ("load_balancing" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("group")
        )
        (
            include.add("snc_name")
            if (
                self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_application_server" in str(self.connection_type.value)
                    )
                    or ("snc_application_server" in str(self.connection_type))
                )
                or self.connection_type
                and (
                    (
                        hasattr(self.connection_type, "value")
                        and self.connection_type.value
                        and "snc_load_balancing" in str(self.connection_type.value)
                    )
                    or ("snc_load_balancing" in str(self.connection_type))
                )
            )
            else exclude.add("snc_name")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)
